from random import randint
from typing import Any, List, Optional, Tuple, Union

import pytest

from json_to_models.dynamic_typing import *


@pytest.mark.xfail(strict=True, raises=ValueError)
def test_metadata_to_typing_with_dict():
    assert metadata_to_typing({'a': 1})


test_imports_compiler_data = [
    pytest.param(
        [
            ('typing', ('List', 'Any')),
            [],
            [],
            ('pytest', 'param'),
            ('typing', ('List', 'Tuple')),
        ],
        "from pytest import param\n"
        "from typing import Any, List, Tuple",
        id="basic"
    ),
    pytest.param(
        [
            ('typing', ('List', 'Any')),
            [],
            ('typing', None),
            [],
            ('pytest', 'param'),
            ('typing', ('List', 'Tuple')),
            ('attr', None),
            ('typing', None),
        ],
        "import attr\n"
        "import typing\n"
        "from pytest import param\n"
        "from typing import Any, List, Tuple",
        id="basic"
    ),
]


@pytest.mark.parametrize("value,expected", test_imports_compiler_data)
def test_imports_compiler(value: ImportPathList, expected):
    code = compile_imports(value)
    assert code == expected


def model(data: dict, name: str):
    meta = ModelMeta(data, str(randint(0, 1000)))
    meta.set_raw_name(name)
    return ModelPtr(meta)


# MetaData | Tuple[import_stmnt, type]
test_data = [
    pytest.param(
        Unknown,
        ('from typing import Any', Any),
        id="UnknownType"
    ),
    pytest.param(
        int,
        ('', int),
        id='builtin'
    ),
    pytest.param(
        DOptional(int),
        ('from typing import Optional', Optional[int]),
        id="DOptional"
    ),
    pytest.param(
        DOptional(Unknown),
        ('from typing import Any, Optional', Optional[Any]),
        id="DOptional_UnknownType"
    ),
    pytest.param(
        DUnion(int, DOptional(str)),
        ('from typing import Optional, Union', Union[int, Optional[str]]),
        id="DUnion"
    ),
    pytest.param(
        DTuple(int, DUnion(str, float)),
        ('from typing import Tuple, Union', Tuple[int, Union[str, float]]),
        id="DTuple"
    ),
    pytest.param(
        DList(DUnion(str, float)),
        ('from typing import List, Union', List[Union[str, float]]),
        id="DList"
    ),
    pytest.param(
        DList(DList(DList(DList(DUnion(str, float))))),
        ('from typing import List, Union', List[List[List[List[Union[str, float]]]]]),
        id="deep_DList"
    ),
    pytest.param(
        FloatString,
        ('from json_to_models.dynamic_typing import FloatString', FloatString),
        id="string_serializable"
    ),
    pytest.param(
        DOptional(IntString),
        ('from json_to_models.dynamic_typing import IntString\n'
         'from typing import Optional', Optional[IntString]),
        id="complex_string_serializable"
    ),
    pytest.param(
        model({'a': int}, "TestModel"),
        ('', 'TestModel')
    ),
    pytest.param(
        DOptional(model({'a': int}, "TestModel")),
        ('from typing import Optional', Optional['TestModel'])
    )
]


@pytest.mark.parametrize("value,expected", test_data)
def test_typing_code_generation(value: MetaData, expected):
    imports, code = metadata_to_typing(value)
    imports_code = compile_imports(imports)
    assert imports_code == expected[0]
    t = eval(code)
    assert t == expected[1]



from typing import Dict, List, Set, Tuple

import pytest

from json_to_models.dynamic_typing import ModelMeta
from json_to_models.generator import MetadataGenerator
from json_to_models.models.structure import compose_models, compose_models_flat, extract_root
from json_to_models.models.utils import ListEx
from json_to_models.registry import ModelRegistry


def test_list_ex():
    l = ListEx(range(10))
    assert l.safe_index('a') is None
    assert l.safe_index(5) == 5
    assert l._safe_indexes(*'abc') == []
    assert l._safe_indexes(*range(20)) == list(range(10))
    l.insert_before('a', 5, 3, 1)
    assert l == [0, 'a', *range(1, 10)]
    l.insert_after('b', 5, 3, 1)
    assert l == [0, 'a', *range(1, 6), 'b', *range(6, 10)]


# This test relies on model names as a some sort of models ids
# and may fail if some logic of their generation will be changed
# List of Tuple[root_model_name, JSON data] | Dict[model_name, Set[root_model_names]]
test_extract_root_data = [
    pytest.param(
        [
            ("TestModelA", {
                "count": 1,
                "items": [
                    {
                        "x": .5,
                        "y": .1
                    }
                ]
            }),
            ("TestModelB", {
                "next": "some_url",
                "prev": None,
                "count": 2000,
                "items": [
                    {
                        "x": .5,
                        "y": .1
                    }
                ]
            }),
        ],
        {
            'Item': {'TestModelA', 'TestModelB'},
            'TestModelA': set()
        }
    ),
    pytest.param(
        [
            ("TestModelA", {
                "count": 1,
                "items": [
                    {
                        "x": .5,
                        "y": .1
                    }
                ]
            }),
            ("TestModelB", {
                "count": 1,
                "items": [
                    {
                        "x": .5,
                        "y": .1
                    }
                ]
            }),
        ],
        {
            'Item': {'TestModelA_TestModelB'},
            'TestModelA_TestModelB': set()
        },
        id="merge_root"
    )
]


@pytest.mark.parametrize("value,expected", test_extract_root_data)
def test_extract_root(models_generator: MetadataGenerator, models_registry: ModelRegistry,
                      value: List[Tuple[str, dict]], expected: Dict[str, Set[str]]):
    for model_name, data in value:
        fields = models_generator.generate(data)
        models_registry.process_meta_data(fields, model_name=model_name)
    models_registry.merge_models(models_generator)
    models_registry.generate_names()
    names_map = {model.index: model.name for model in models_registry.models}
    names_map.update({model.name: model.index for model in models_registry.models})
    for model_name, roots in expected.items():
        meta = models_registry.models_map[names_map[model_name]]
        extracted_roots = {names_map[ix] for ix in extract_root(meta)}
        assert extracted_roots == roots


base_dict = {"field_" + str(i): int for i in range(20)}


def _test_compose_models(
        function, request,
        models_generator: MetadataGenerator, models_registry: ModelRegistry,
        value: List[Tuple[str, dict]], expected: List[Tuple[str, list]], expected_mapping: Dict[str, str]
):
    for model_name, metadata in value:
        models_registry.process_meta_data(metadata, model_name=model_name)
    models_registry.merge_models(models_generator)
    models_registry.generate_names()
    names_map = {model.index: model.name for model in models_registry.models}
    names_map.update({model.name: model.index for model in models_registry.models})
    root, mapping = function(models_registry.models_map)

    def check(nested_value: List[dict], nested_expected: List[Tuple[str, list]]):
        for model_dict, (model_name, nested) in zip(nested_value, nested_expected):
            assert model_dict["model"].name == model_name, str(nested_value)
            assert len(model_dict["nested"]) == len(nested), f"(Parent model is {model_name})"
            check(model_dict["nested"], nested)

    check(root, expected)

    name = lambda model: model.name if isinstance(model, ModelMeta) else model
    mapping = {name(model): name(parent) for model, parent in mapping.items()}
    assert mapping == expected_mapping


# This test relies on model names as a some sort of models ids
# List of Tuple[root_model_name, model-meta] | List[Tuple[model_name, nested_models]]]
# where nested_models is a recursive definition
test_compose_models_data = [
    pytest.param(
        [
            ("A", {
                "field1": int,
                "item": {
                    "another_field": float
                }
            })
        ],
        [
            ("A", [
                ("Item", [])
            ])
        ],
        {},
        id="basic_test"
    ),
    pytest.param(
        [
            ("RootA", {
                "item": {
                    "field": float
                }
            }),
            ("RootB", {
                "item": {
                    "field": float
                },
                "idontwantrootmodelstomerge": bool
            })
        ],
        [
            ("Item", []),
            ("RootA", []),
            ("RootB", [])
        ],
        {},
        id="global_nested_model"
    ),
    pytest.param(
        [
            ("RootA", {
                "item": {
                    "field": float
                }
            }),
            ("RootB", {
                "item": {
                    "field": float
                }
            })
        ],
        [
            ("RootA_RootB", [
                ("Item", [])
            ])
        ],
        {},
        id="roots_merge"
    ),
    pytest.param(
        [
            ("RootFirst", {
                "root_field": float
            }),
            ("RootA", {
                "item": {
                    "field": float
                }
            }),
            ("RootB", {
                "item": {
                    "field": float
                },
                "idontwantrootmodelstomerge": bool
            })
        ],
        [
            ("RootFirst", []),
            ("Item", []),
            ("RootA", []),
            ("RootB", [])
        ],
        {},
        id="root_order"
    ),
    pytest.param(
        [
            ("Root", {
                "model_a": {
                    "field_a": {
                        "field": float
                    }
                },
                "model_b": {
                    "field_b": {
                        "field": float
                    }
                }
            }),
        ],
        [
            ("Root", [
                ("FieldA_FieldB", []),
                ("ModelA", []),
                ("ModelB", []),
            ])
        ],
        {'FieldA_FieldB': 'Root'},
        id="generic_in_nested_models"
    ),
    pytest.param(
        [
            ("RootItem", {
                "field": float
            }),
            ("RootA", {
                "item": {
                    "field": float
                },
                **base_dict
            }),
            ("RootB", base_dict)
        ],
        [
            ("RootItem", []),
            ("RootA_RootB", [])
        ],
        {},
        id="merge_with_root_model"
    ),
    pytest.param(
        [
            ("Root", {
                "model_a": {
                    "field_a": {
                        "field": {
                            "nested_field": float
                        }
                    }
                },
                "model_b": {
                    "field_b": {
                        "field": {
                            "nested_field": float
                        }
                    }
                }
            }),
        ],
        [
            ("Root", [
                ("FieldA_FieldB", [
                    ("Field", [])
                ]),
                ("ModelA", []),
                ("ModelB", []),
            ])
        ],
        {'FieldA_FieldB': 'Root'},
        id="generic_in_nested_models_with_nested_model"
    ),
]


@pytest.mark.parametrize("value,expected,expected_mapping", test_compose_models_data)
def test_compose_models(
        request,
        models_generator: MetadataGenerator, models_registry: ModelRegistry,
        value: List[Tuple[str, dict]], expected: List[Tuple[str, list]], expected_mapping: Dict[str, str]
):
    _test_compose_models(compose_models, request, models_generator, models_registry, value, expected, expected_mapping)


test_compose_models_flat_data = [
    pytest.param(
        [
            ("A", {
                "field1": int,
                "item": {
                    "another_field": float
                }
            })
        ],
        [
            ("A", []),
            ("Item", [])
        ],
        {},
        id="basic_test"
    ),
    pytest.param(
        [
            ("RootA", {
                "item": {
                    "field": float
                }
            }),
            ("RootB", {
                "item": {
                    "field": float
                },
                "idontwantrootmodelstomerge": bool
            })
        ],
        [
            ("RootA", []),
            ("RootB", []),
            ("Item", []),
        ],
        {},
        id="global_nested_model"
    ),
    pytest.param(
        [
            ("RootA", {
                "item": {
                    "field": float
                }
            }),
            ("RootB", {
                "item": {
                    "field": float
                }
            })
        ],
        [
            ("RootA_RootB", []),
            ("Item", [])
        ],
        {},
        id="roots_merge"
    ),
    pytest.param(
        [
            ("RootFirst", {
                "root_field": float
            }),
            ("RootA", {
                "item": {
                    "field": float
                }
            }),
            ("RootB", {
                "item": {
                    "field": float
                },
                "idontwantrootmodelstomerge": bool
            })
        ],
        [
            ("RootFirst", []),
            ("RootA", []),
            ("RootB", []),
            ("Item", []),
        ],
        {},
        id="root_order"
    ),
    pytest.param(
        [
            ("Root", {
                "model_a": {
                    "field_a": {
                        "field": float
                    }
                },
                "model_b": {
                    "field_b": {
                        "field": float
                    }
                }
            }),
        ],
        [
            ("Root", []),
            ("ModelA", []),
            ("ModelB", []),
            ("FieldA_FieldB", []),

        ],
        {},
        id="generic_in_nested_models"
    ),
    pytest.param(
        [
            ("RootItem", {
                "field": float
            }),
            ("RootA", {
                "item": {
                    "field": float
                },
                **base_dict
            }),
            ("RootB", base_dict)
        ],
        [
            ("RootA_RootB", []),
            ("RootItem", []),
        ],
        {},
        id="merge_with_root_model"
    ),
    pytest.param(
        [
            ("Root", {
                "model_a": {
                    "field_a": {
                        "field": {
                            "nested_field": float
                        }
                    }
                },
                "model_b": {
                    "field_b": {
                        "field": {
                            "nested_field": float
                        }
                    }
                }
            }),
        ],
        [
            ("Root", []),
            ("ModelA", []),
            ("ModelB", []),
            ("FieldA_FieldB", []),
            ("Field", [])
        ],
        {},
        id="generic_in_nested_models_with_nested_model"
    ),
    pytest.param(
        [
            ("RootA", {
                "field_a1": {"nested": {"field": int}},
                "field_a2": {"nested_2": {"field_2": int}},
                "nested": {"field": int}
            }),
            ("RootB", {
                "field_b1": {"nested": {"field": int}},
                "field_b2": {"nested_2": {"field_2": int}},
                "nested": {"field": int}
            }),
            ("RootC", {
                "field_c1": float
            }),
        ],
        [
            ("RootA", []),
            ("RootB", []),
            ("RootC", []),
            ("FieldA1_FieldB1", []),
            ("Nested", []),
            ("FieldA2_FieldB2", []),
            ("Nested2", []),
        ],
        {},
        id="sort_1"
    ),
    pytest.param(
        [
            ("RootA", {
                "field_a1": {"field_a1_n": {"field": int}},
                "field_a2": {"field_a2_n": {"field": int}},
                "field_a3": {"field_a3_n": {"field": int}},
                "field_a4": {"field_a4_n": {"field": int}},
            }),
            ("RootB", {
                "field_b1": {"field_b1_n": {"field": int}},
                "field_b2": {"field_b2_n": {"field": int}},
                "field_b3": {"field_b3_n": {"field": int}},
                "field_b4": {"field_b4_n": {"field": int}},
            }),
            ("RootC", {
                "field_c1": {"field_c1_n": {"field": int}},
                "field_c2": {"field_c2_n": {"field": int}},
                "field_c3": {"field_c3_n": {"field": int}},
                "field_c4": {"field_c4_n": {"field": int}},
            }),
        ],
        [
            ("RootA", []),
            ("RootB", []),
            ("RootC", []),
            ("FieldA1", []),
            ("FieldA2", []),
            ("FieldA3", []),
            ("FieldA4", []),
            ("FieldB1", []),
            ("FieldB2", []),
            ("FieldB3", []),
            ("FieldB4", []),
            ("FieldC1", []),
            ("FieldC2", []),
            ("FieldC3", []),
            ("FieldC4", []),
            ("FieldA1N_FieldA2N_FieldA3N_FieldA4N_"
             "FieldB1N_FieldB2N_FieldB3N_FieldB4N_"
             "FieldC1N_FieldC2N_FieldC3N_FieldC4N", []),
        ],
        {},
        id="sort_2"
    ),
    pytest.param(
        [
            ("RootA", {
                "field_a1": {
                    "field_a1_n": {"field_1": int},
                    "field_a1_n2": {"field_11": int},
                },
                "field_a2": {"field_a2_n": {"field_2": int}},
                "field_a3": {"field_a3_n": {"field_3": int}},
                "field_a4": {"field_a4_n": {"field_4": int}},
            }),
            ("RootB", {
                "field_b1": {
                    "field_b1_n": {"field_01": int},
                    "field_b1_n2": {"field_b11": int},
                    "field_b1_n3": {"field_b21": int},
                },
                "field_b2": {"field_b2_n": {"field_02": int}},
                "field_b3": {
                    "field_b3_n": {"field_03": int},
                    "field_b3_n2": {"field_b13": int},
                    "field_b3_n3": {"field_b23": int},
                    "field_b3_n4": {"field_b33": int},
                },
                "field_b4": {"field_b4_n": {"field_04": int}},
            }),
        ],
        [
            ("RootA", []),
            ("RootB", []),

            ("FieldA1", []),
            ("FieldA1N", []),
            ("FieldA1N2", []),

            ("FieldA2", []),
            ("FieldA2N", []),

            ("FieldA3", []),
            ("FieldA3N", []),

            ("FieldA4", []),
            ("FieldA4N", []),

            ("FieldB1", []),
            ("FieldB1N", []),
            ("FieldB1N2", []),
            ("FieldB1N3", []),

            ("FieldB2", []),
            ("FieldB2N", []),

            ("FieldB3", []),
            ("FieldB3N", []),
            ("FieldB3N2", []),
            ("FieldB3N3", []),
            ("FieldB3N4", []),

            ("FieldB4", []),
            ("FieldB4N", []),
        ],
        {},
        id="sort_3"
    ),
]


@pytest.mark.parametrize("value,expected,expected_mapping", test_compose_models_flat_data)
def test_compose_models_flat(
        request,
        models_generator: MetadataGenerator, models_registry: ModelRegistry,
        value: List[Tuple[str, dict]], expected: List[Tuple[str, list]], expected_mapping: Dict[str, str]
):
    _test_compose_models(compose_models_flat, request, models_generator, models_registry, value, expected,
                         expected_mapping)



import pytest

from json_to_models.dynamic_typing import IsoTimeString
from json_to_models.dynamic_typing.string_serializable import (FloatString, IntString, StringSerializable,
                                                               StringSerializableRegistry)
from json_to_models.generator import MetadataGenerator

r = StringSerializableRegistry()


@r.add()
class A(StringSerializable):
    pass


@r.add({A})
class D(StringSerializable):
    pass


@r.add({D, A})
class E(StringSerializable):
    pass


@r.add({A, E, D})
class B(StringSerializable):
    pass


@r.add({B, A, E, D})
class C(StringSerializable):
    pass


@r.add({C, B, A, E, D})
class F(StringSerializable):
    pass


@r.add({A})
class X(StringSerializable):
    pass


@r.add({X})
class Y(StringSerializable):
    pass


# resolve args | resolved nodes
test_data = [
    pytest.param((A, B), {B}),
    pytest.param((A, B, C), {C}),
    pytest.param((B, B), {B}),
    pytest.param((A, B, D), {B}),
    pytest.param((X, B), {B, X}),
]


@pytest.mark.parametrize("value,expected", test_data)
def test_string_serializable_registry(value, expected):
    result = r.resolve(*value)
    assert result == expected


def test_string_serializable_registry_order():
    r2 = StringSerializableRegistry()
    gen = MetadataGenerator(r2)

    r2.add(cls=IsoTimeString)
    r2.add(cls=IntString)
    r2.add(replace_types=(IntString,), cls=FloatString)
    assert gen._detect_type("12") != IntString

    r2.remove(IsoTimeString)
    r2.add(cls=IsoTimeString)
    assert gen._detect_type("12") == IntString
    assert gen._detect_type("12:14") == IsoTimeString



from random import shuffle

import pytest
from inflection import singularize

from json_to_models.utils import (Index, cached_classmethod, cached_method, convert_args_decorator, distinct_words,
                                  json_format)

test_distinct_words_data = [
    pytest.param(['test', 'foo', 'bar'], {'test', 'foo', 'bar'}),
    pytest.param(['test', 'foo', 'bar', 'foo_bar'], {'test', 'foo', 'bar'}),
    pytest.param(['awesome_asdawdaw', 'awesome_testing', 'test', 'simple_test', 'awesome'], {'test', 'awesome'}),
    pytest.param(['awesome_testing', 'test'], {'test'}),
]


@pytest.mark.parametrize("value,expected", test_distinct_words_data)
def test_distinct_words(value, expected):
    shuffled = value[:]
    shuffle(shuffled)
    for v in (value, reversed(value), shuffled):
        v = list(v)
        words_set = distinct_words(*v)
        assert words_set == expected, f"Test case: {v}"


def test_json_format():
    d = {'a': 2, 'aswr': [*range(10)]}
    d['d'] = d.copy()
    s = json_format(d)
    print(s)
    assert '\n' in s
    assert ' ' * 4 in s


def test_singularize():
    data = {
        'dogs': 'dog',
        'properties': 'property',
        'cat': 'cat'
    }
    for plur, sing in data.items():
        assert singularize(plur) == sing


def test_index():
    ix = Index()
    for _ in range(1000):
        ix()


@convert_args_decorator(int, b=float)
def f(a, b):
    return a + b


class A:
    @convert_args_decorator(float, float, method=True)
    def __init__(self, x, y):
        self.value = x + y


def test_convert_args_decorator():
    assert f('1', b='1.5') == 2.5
    a = A("2.3", "7.5")
    assert a.value == 9.8


def test_cached_methods():
    class A:
        x = []

        def __init__(self):
            self.y = []

        @cached_method
        def f(self, a):
            self.y.append(a)
            return a

        @cached_classmethod
        def g(cls, a):
            cls.x.append(a)
            return a

    A.g('a')
    A.g('a')
    a = A()
    A.g('b')
    A.g('a')
    a.f('b')
    a.f('b')
    a.f('a')
    a.f('a')
    b = A()
    a.f('b')
    b.f('a')
    b.g('c')

    assert A.x == ['a', 'b', 'c']
    assert a.y == ['b', 'a']
    assert b.y == ['a']
    assert a.x == ['a', 'b', 'c']



import json
from functools import wraps
from typing import Callable, Optional, Set


class Index:
    def __init__(self):
        self.ch = 'A'
        self.i = 1

    def __call__(self, *args, **kwargs):
        value = f'{self.i}{self.ch}'
        ch = chr(ord(self.ch) + 1)
        if ch <= 'Z':
            self.ch = ch
        else:
            self.ch = 'A'
            self.i += 1
        return value


def json_format(x) -> str:
    return json.dumps(x, indent=4, default=str, ensure_ascii=False)


def distinct_words(*words: str) -> Set[str]:
    """
    Filters strings so only unique strings without extended ones will be exists in resulted set, e.g.
    >>> distinct_words('test', 'another_test', 'foo', 'qwerty_foo_bar')
    {'test', 'foo'}

    :param words:
    :return:
    """
    words = set(words)
    filtered_words = set()
    for name in words:
        flag = True
        for other in list(filtered_words):
            if name in other:
                filtered_words.add(name)
                filtered_words.remove(other)
                flag = False
            elif other in name:
                flag = False
        if flag:
            filtered_words.add(name)
    return filtered_words


def convert_args(fn: Callable, *args_converters: Optional[type], **kwargs_converters: Optional[type]) -> Callable:
    """
    Decorator. Apply ``args_converters`` to callable arguments and kwargs_converters to kwargs.
    If converter is None then argument will passed as is.

    :param fn: Function or class
    :param args_converters: Arguments converters
    :param kwargs_converters: Keyword arguments converters
    :return: Callable wrapper
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        converted = (
            t(value) if t else value
            for value, t in zip(args, args_converters)
        )
        kwargs_converted = {
            name: kwargs_converters[name](kwargs[name]) if kwargs_converters.get(name, None) else kwargs[name]
            for name in kwargs.keys()
        }
        if len(args_converters) < len(args):
            remain = args[len(args_converters):]
        else:
            remain = ()
        return fn(*converted, *remain, **kwargs_converted)

    return wrapper


def convert_args_decorator(*args_converters: type, method=False, **kwargs_converters):
    """
    Decorator factory.

    :param args_converters: Arguments converters
    :param method: Set to True if decorated function is method or classmethod
    :param kwargs_converters: Keyword arguments converters
    :return:
    """

    def decorator(fn):
        if method:
            return convert_args(fn, None, *args_converters, **kwargs_converters)
        else:
            return convert_args(fn, *args_converters, **kwargs_converters)

    return decorator


def cached_method(func: Callable):
    """
    Decorator to cache method return values
    """

    @wraps(func)
    def cached_fn(self, *args):
        if getattr(self, '__cache__', None) is None:
            setattr(self, '__cache__', {})
        value = self.__cache__.get(args, ...)
        if value is Ellipsis:
            value = func(self, *args)
            self.__cache__[args] = value
        return value

    return cached_fn


def cached_classmethod(func: Callable):
    """
    Decorator to cache classmethod return values
    """
    cache = {}

    @wraps(func)
    def cached_fn(cls, *args):
        value = cache.get(args, ...)
        if value is Ellipsis:
            value = func(cls, *args)
            cache[args] = value
        return value

    return classmethod(cached_fn)



from typing import Iterable

import pytest

from json_to_models.dynamic_typing import DList, DOptional, ModelMeta, Unknown
from json_to_models.generator import MetadataGenerator
from json_to_models.registry import ModelRegistry
from test.test_registry.test_registry_process_meta_data import check_type, cycle_ref, test_data as base_test_data

# Include test cases from `test_registry_process_meta_data.py`. They should be correct.
# At least until their not contains any models with same keys.
# Original MetaData | Merged models
test_data = base_test_data + [
    pytest.param(
        [
            {
                'a': int,
                'b': float
            },
            {
                'a': int,
                'b': float
            }
        ],
        [
            {
                'a': int,
                'b': float
            }
        ],
        id="simple_merge"
    ),
    pytest.param(
        [
            {
                'd': int,
                'e': float
            },
            {
                'a': int,
                'b': float
            },
            {
                'a': int,
                'b': float
            }
        ],
        [
            {
                'd': int,
                'e': float
            },
            {
                'a': int,
                'b': float
            }
        ],
        id="simple_merge_with_other_model"
        # Also test sort_models by this test case
    ),
    pytest.param(
        [
            {
                'a': int,
                'b': float
            },
            {
                'd': int,
                'e': {
                    'a': int,
                    'b': float
                },
            }
        ],
        [
            {
                'a': int,
                'b': float
            },
            {
                'd': int,
                'e': {
                    'a': int,
                    'b': float
                },
            }
        ],
        id="merge_nested"
    ),
    pytest.param(
        {
            'int': int,
            'items': DList({
                'int': int,
                'items': DList(Unknown)  # Empty list
            })
        },
        [
            {
                'int': int,
                'items': DList({
                    'int': int,
                    'items': DList(cycle_ref)
                })
            }
        ],
        id="cycle"
    ),
    pytest.param(
        {
            'int': int,
            'items': {
                'count': int,
                'wrapper': DList({
                    'int': int,
                    'items': Unknown  # null
                })
            }
        },
        [
            {
                'int': int,
                'items': {
                    'count': int,
                    'wrapper': DList({
                        'int': int,
                        'items': {
                            'count': int,
                            'wrapper': DList(cycle_ref)
                        }
                    })
                }
            },
            {
                'count': int,
                'wrapper': DList({
                    'int': int,
                    'items': {
                        'count': int,
                        'wrapper': DList(cycle_ref)
                    }
                })
            }
        ],
        id="cycle_with_wrapper"
    ),
    pytest.param(
        [
            {
                "field" + str(i): int for i in range(10)
            },
            {
                "field" + str(i): int for i in range(1, 10)
            },
            {
                "field" + str(i): int for i in range(9)
            }
        ],
        [
            {
                "field0": DOptional(int),
                **{
                    "field" + str(i): int for i in range(1, 9)
                },
                "field9": DOptional(int),
            }
        ],
        id="merge_models"
    ),
    pytest.param(
        [
            {
                "field" + str(i): int for i in range(10)
            },
            {
                "field" + str(i): int for i in range(5, 10)
            },
            {
                "field" + str(i): int for i in range(5)
            }
        ],
        [
            {
                "field" + str(i): int for i in range(10)
            },
            {
                "field" + str(i): int for i in range(5, 10)
            },
            {
                "field" + str(i): int for i in range(5)
            }
        ],
        id="not_merge_models"
    ),
    pytest.param(
        [
            {
                "field" + str(i): int for i in range(10)
            },
            {
                "field" + str(i): (int if i < 9 else DOptional(int)) for i in range(10)
            },
            {
                "field" + str(i): int for i in range(9)
            }
        ],
        [
            {
                **{
                    "field" + str(i): int for i in range(9)
                },
                "field9": DOptional(int),
            }
        ],
        id="merge_models_with_optional_field"
    ),
    pytest.param(
        [
            {
                "a" + str(i): int for i in range(20)
            },
            {
                **{
                    "a" + str(i): int for i in range(20)
                },
                **{
                    "b" + str(i): int for i in range(20)
                }
            },
            {
                "b" + str(i): int for i in range(20)
            },
            {
                "c" + str(i): int for i in range(20)
            },
            {
                **{
                    "b" + str(i): int for i in range(20)
                },
                **{
                    "c" + str(i): int for i in range(20)
                }
            },
            {
                "field1": int
            },
            {
                "field1": int
            }
        ],
        [
            {
                **{
                    "a" + str(i): DOptional(int) for i in range(20)
                },
                **{
                    "b" + str(i): DOptional(int) for i in range(20)
                },
                **{
                    "c" + str(i): DOptional(int) for i in range(20)
                }
            },
            {
                "field1": int
            }
        ],
        id="multistage_merge"
    ),
]


def sort_models(models: Iterable[ModelMeta]):
    return sorted(
        models,
        key=lambda model: ",".join(sorted(
            (model if isinstance(model, dict) else model.type).keys()
        ))
    )


@pytest.mark.parametrize("value,expected", test_data)
def test_registry_merge_models_base(models_generator: MetadataGenerator, models_registry: ModelRegistry, value,
                                    expected):
    if isinstance(value, list):
        for v in value:
            models_registry.process_meta_data(v)
    else:
        models_registry.process_meta_data(value)
    models_registry.merge_models(generator=models_generator)
    assert len(models_registry.models) == len(expected)
    for model, expected_model in zip(sort_models(models_registry.models), sort_models(expected)):
        check_type(model, expected_model)



import re
from typing import Any, Callable, List, Optional, Pattern, Union

from .dynamic_typing import (
    ComplexType,
    DDict,
    DList,
    DOptional,
    DUnion,
    MetaData,
    ModelPtr,
    Null,
    SingleType,
    StringLiteral,
    StringSerializable,
    StringSerializableRegistry,
    Unknown,
    registry
)

_static_types = {float, bool, int}


class MetadataGenerator:
    CONVERTER_TYPE = Optional[Callable[[str], Any]]

    def __init__(
            self,
            str_types_registry: StringSerializableRegistry = None,
            dict_keys_regex: List[Union[Pattern, str]] = None,
            dict_keys_fields: List[str] = None
    ):
        """

        :param str_types_registry: StringSerializableRegistry instance. Default registry will be used if None passed .
        :param dict_keys_regex: List of RegExpressions (compiled or not).
            If all keys of some dict are match one of them then this dict will be marked as dict field
            but not nested model.
        :param dict_keys_fields: List of model fields names that will be marked as dict field
        """
        self.str_types_registry = str_types_registry if str_types_registry is not None else registry
        self.dict_keys_regex = [re.compile(r) for r in dict_keys_regex] if dict_keys_regex else []
        self.dict_keys_fields = set(dict_keys_fields or ())

    def generate(self, *data_variants: dict) -> dict:
        """
        Convert given list of data variants to metadata dict
        """
        fields_sets = [self._convert(data) for data in data_variants]
        fields = self.merge_field_sets(fields_sets)
        return self.optimize_type(fields)

    def _convert(self, data: dict):
        """
        Key and string value converting
        """
        fields = dict()
        for key, value in data.items():
            if not isinstance(key, str):
                raise TypeError(f'You probably using some not JSON-compatible parser and have some {type(key)} as dict key. '
                                f'This is not supported.\n'
                                f'Context: {data}\n'
                                f'(If you parsing yaml try to replace PyYaml with ruamel.yaml)')
            convert_dict = key not in self.dict_keys_fields
            fields[key] = self._detect_type(value, convert_dict)
        return fields

    def _detect_type(self, value, convert_dict=True) -> MetaData:
        """
        Converts json value to metadata
        """
        # Simple types
        t = type(value)
        if t in _static_types:
            return t

        # List trying to yield nested type
        elif t is list:
            if value:
                types = [self._detect_type(item) for item in value]
                if len(types) > 1:
                    union = DUnion(*types)
                    if len(union.types) == 1:
                        return DList(*union.types)
                    return DList(union)
                else:
                    return DList(*types)
            else:
                return DList(Unknown)

        # Dict should be processed as another model if convert_dict is enabled
        elif isinstance(value, dict):
            if not value:
                return DDict(Unknown)
            for reg in self.dict_keys_regex:
                if all(map(reg.match, value.keys())):
                    convert_dict = False
                    break

            if convert_dict:
                return self._convert(value)
            else:
                types = [self._detect_type(item) for item in value.values()]
                if len(types) > 1:
                    union = DUnion(*types)
                    if len(union.types) == 1:
                        return DDict(*union.types)
                    return DDict(union)
                else:
                    return DDict(*types)

        # null interpreted as is and will be processed later on Union merge stage
        elif value is None:
            return Null

        # string types trying to convert to other string-serializable types
        else:
            for t in self.str_types_registry:
                try:
                    value = t.to_internal_value(value)
                except ValueError:
                    continue
                return t
            return StringLiteral({value})

    def merge_field_sets(self, field_sets: List[MetaData]) -> MetaData:
        """
        Merge fields sets into one set of pairs (key, metadata)
        """
        fields: dict = {}

        first = True
        for model in field_sets:
            fields_diff = set(fields.keys())

            for name, field in model.items():
                if name not in fields:
                    # New field
                    field = field if first or isinstance(field, DOptional) else DOptional(field)
                else:
                    field_original = fields[name]
                    fields_diff.remove(name)
                    if isinstance(field_original, DOptional):
                        # Existing optional field
                        if field_original == field or field_original.type == field:
                            continue
                        field_original = field_original.type
                        field = DOptional(DUnion(
                            *(field.types if isinstance(field, DUnion) else [field]),
                            *(field_original.types if isinstance(field_original, DUnion) else [field_original])
                        ))
                        if len(field.type) == 1:
                            field.type = field.type.types[0]
                    else:
                        if field_original == field or (isinstance(field, DOptional) and field_original == field.type):
                            continue
                        field = DUnion(
                            *(field.types if isinstance(field, DUnion) else [field]),
                            *(field_original.types if isinstance(field_original, DUnion) else [field_original])
                        )
                        if len(field) == 1:
                            field = field.types[0]

                fields[name] = field

            for name in fields_diff:
                # Missing fields becomes optionals
                if not isinstance(fields[name], DOptional):
                    fields[name] = DOptional(fields[name])

            first = False
        return fields

    def optimize_type(self, meta: MetaData, process_model_ptr=False) -> MetaData:
        """
        Finds some redundant types and replace them with a simpler one

        :param process_model_ptr: Control whether process ModelPtr instances or not.
            Default is False to prevent recursion cycles.
        """
        if isinstance(meta, dict):
            fields = {}

            for k, v in meta.items():
                fields[k] = self.optimize_type(v)
            return fields

        elif isinstance(meta, DUnion):
            return self._optimize_union(meta)

        elif isinstance(meta, DOptional):
            t = self.optimize_type(meta.type)
            if isinstance(t, DOptional):
                t = t.type
            return meta.replace(t)

        elif isinstance(meta, SingleType) and (process_model_ptr or not isinstance(meta, ModelPtr)):
            # Optimize nested type
            return meta.replace(self.optimize_type(meta.type))

        elif isinstance(meta, ComplexType):
            # Optimize all nested types
            return meta.replace([self.optimize_type(nested) for nested in meta])
        elif isinstance(meta, StringLiteral):
            if meta.overflowed or not meta.literals:
                return str
        return meta

    def _optimize_union(self, t: DUnion):
        # Replace DUnion of 1 element with this element
        # if len(t) == 1:
        #     return t.types[0]

        # Split nested types into categories
        str_types: List[Union[type, StringSerializable]] = []
        types_to_merge: List[dict] = []
        list_types: List[DList] = []
        dict_types: List[DDict] = []
        other_types: List[MetaData] = []
        for item in t.types:
            if isinstance(item, DOptional):
                item = item.type
                other_types.append(Null)
            if isinstance(item, dict):
                types_to_merge.append(item)
            elif item in self.str_types_registry or item is str:
                str_types.append(item)
            elif isinstance(item, DList):
                list_types.append(item)
            elif isinstance(item, DDict):
                dict_types.append(item)
            else:
                other_types.append(item)

        if int in other_types and float in other_types:
            other_types.remove(int)

        if types_to_merge:
            other_types.append(self.merge_field_sets(types_to_merge))

        for cls, iterable_types in ((DList, list_types), (DDict, dict_types)):
            if iterable_types:
                other_types.append(cls(DUnion(*(
                    t.type for t in iterable_types
                ))))

        if str in str_types:
            other_types.append(str)
        elif str_types:
            str_types = self.str_types_registry.resolve(*str_types)
            # Replace str pseudo-types with <class 'str'> when they can not be resolved into single type
            other_types.append(str if len(str_types) > 1 else next(iter(str_types)))

        types = [self.optimize_type(t) for t in other_types]


        if len(types) > 1:
            if Unknown in types:
                types.remove(Unknown)

            optional = False
            if Null in types:
                optional = True
                while Null in types:
                    types.remove(Null)

            meta_type = DUnion(*types)
            if len(meta_type.types) == 1:
                meta_type = meta_type.types[0]

            if optional:
                return DOptional(meta_type)
        else:
            meta_type = types[0]

        return meta_type



from typing import Dict, Iterable, List, Set, Tuple

from . import Index, ModelsStructureType
from .utils import ListEx, PositionsDict
from ..dynamic_typing import BaseType, DOptional, ModelMeta, ModelPtr


def compose_models(models_map: Dict[str, ModelMeta]) -> ModelsStructureType:
    """
    Generate nested sorted models structure for internal usage.

    :return: List of root models data, Map(child model -> root model) for absolute ref generation
    """
    root_models = ListEx()
    root_nested_ix = 0
    structure_hash_table: Dict[Index, dict] = {
        key: {
            "model": model,
            "nested": ListEx(),
            "roots": list(extract_root(model)),  # Indexes of root level models
        } for key, model in models_map.items()
    }
    # TODO: Test path_injections
    path_injections: Dict[ModelMeta, ModelMeta] = {}

    for key, model in models_map.items():
        pointers = list(filter_pointers(model))
        has_root_pointers = len(pointers) != len(model.pointers)
        if not pointers:
            # Root level model
            if not has_root_pointers:
                raise Exception(f'Model {model.name} has no pointers')
            root_models.append(structure_hash_table[key])
        else:
            parents = {ptr.parent.index for ptr in pointers}
            struct = structure_hash_table[key]
            # Model is using by other models
            if has_root_pointers or len(parents) > 1 and len(struct["roots"]) > 1:
                # Model is using by different root models
                try:
                    root_models.insert_before(
                        struct,
                        *(structure_hash_table[parent_key] for parent_key in struct["roots"])
                    )
                except ValueError:
                    root_models.insert(root_nested_ix, struct)
                    root_nested_ix += 1
            elif len(parents) > 1 and len(struct["roots"]) == 1:
                # Model is using by single root model
                parent = structure_hash_table[struct["roots"][0]]
                parent["nested"].insert(0, struct)
                path_injections[struct["model"]] = parent["model"]
            else:
                # Model is using by only one model
                parent = structure_hash_table[next(iter(parents))]
                struct = structure_hash_table[key]
                parent["nested"].append(struct)

    return root_models, path_injections


def compose_models_flat(models_map: Dict[Index, ModelMeta]) -> ModelsStructureType:
    """
    Generate flat sorted (by nesting level, ASC) models structure for internal usage.

    :param models_map: Mapping (model index -> model meta instance).
    :return: List of root models data, Map(child model -> root model) for absolute ref generation
    """
    root_models = ListEx()
    positions: PositionsDict[Index, int] = PositionsDict()
    top_level_models: Set[Index] = set()
    structure_hash_table: Dict[Index, dict] = {
        key: {
            "model": model,
            "nested": ListEx(),
            "roots": list(extract_root(model)),  # Indexes of root level models
        } for key, model in models_map.items()
    }

    for key, model in models_map.items():
        pointers = list(filter_pointers(model))
        has_root_pointers = len(pointers) != len(model.pointers)
        if not pointers:
            # Root level model
            if not has_root_pointers:
                raise Exception(f'Model {model.name} has no pointers')
            root_models.insert(positions["root"], structure_hash_table[key])
            top_level_models.add(key)
            positions.update_position("root", PositionsDict.INC)
        else:
            parents = {ptr.parent.index for ptr in pointers}
            struct = structure_hash_table[key]
            # Model is using by other models
            if has_root_pointers or len(parents) > 1 and len(struct["roots"]) >= 1:
                # Model is using by different root models
                if parents & top_level_models:
                    parents.add("root")
                parents_positions = {positions[parent_key] for parent_key in parents
                                     if parent_key in positions}
                parents_joined = "#".join(sorted(parents))
                if parents_joined in positions:
                    parents_positions.add(positions[parents_joined])
                pos = max(parents_positions) if parents_positions else len(root_models)
                positions.update_position(parents_joined, pos + 1)
            else:
                # Model is using by only one model
                parent = next(iter(parents))
                pos = positions.get(parent, len(root_models))
                positions.update_position(parent, pos + 1)
            positions.update_position(key, pos + 1)
            root_models.insert(pos, struct)

    return root_models, {}


def filter_pointers(model: ModelMeta) -> Iterable[ModelPtr]:
    """
    Return iterator over pointers with not None parent
    """
    return (ptr for ptr in model.pointers if ptr.parent)


def extract_root(model: ModelMeta) -> Set[Index]:
    """
    Return set of indexes of root models that are use given `model` directly or through another nested model.
    """
    seen: Set[Index] = set()
    nodes: List[ModelPtr] = list(filter_pointers(model))
    roots: Set[Index] = set()
    while nodes:
        node = nodes.pop()
        seen.add(node.type.index)
        filtered = list(filter_pointers(node.parent))
        nodes.extend(ptr for ptr in filtered if ptr.type.index not in seen)
        if not filtered:
            roots.add(node.parent.index)
    return roots


def sort_fields(model_meta: ModelMeta, unicode_fix=False) -> Tuple[List[str], List[str]]:
    """
    Split fields into required and optional groups

    :return: two list of fields names: required fields, optional fields
    """
    fields = model_meta.type
    required = []
    required_2 = []
    optional = []
    for key, meta in fields.items():
        if isinstance(meta, DOptional):
            optional.append(key)
        elif unicode_fix and isinstance(meta, BaseType) and any(
                isinstance(node, ModelMeta)
                for node in meta.iter_child()
        ):
            required_2.append(key)
        else:
            required.append(key)
    return required + required_2, optional



from collections import defaultdict
from typing import Generic, TypeVar, Union

from . import INDENT

T = TypeVar('T')


class ListEx(list, Generic[T]):
    """
    Extended list with shortcut methods
    """

    def safe_index(self, value: T):
        try:
            return self.index(value)
        except ValueError:
            return None

    def _safe_indexes(self, *values: T):
        return [i for i in map(self.safe_index, values) if i is not None]

    def insert_before(self, value: T, *before: T):
        ix = self._safe_indexes(*before)
        if not ix:
            raise ValueError
        pos = min(ix)
        self.insert(pos, value)
        return pos

    def insert_after(self, value: T, *after: T):
        ix = self._safe_indexes(*after)
        if not ix:
            raise ValueError
        pos = max(ix) + 1
        self.insert(pos, value)
        return pos


class PositionsDict(defaultdict):
    # Dict contains mapping Index -> position, where position is list index to insert nested element of Index
    INC = object()

    def __init__(self, default_factory=int, **kwargs):
        super().__init__(default_factory, **kwargs)

    def update_position(self, key: str, value: Union[object, int]):
        """
        Shift all elements which are placed after updated one

        :param key: Index or "root"
        :param value: Could be position or PositionsDict.INC to perform quick increment (x+=1)
        :return:
        """
        if value is self.INC:
            value = self[key] + 1
        if key in self:
            old_value = self[key]
            delta = value - old_value
        else:
            old_value = value
            delta = 1
        for k, v in self.items():
            if k != key and v >= old_value:
                self[k] += delta
        self[key] = value


def indent(string: str, lvl: int = 1, indent: str = INDENT) -> str:
    """
    Indent all lines of string by ``indent * lvl``
    """
    return "\n".join(indent * lvl + line for line in string.split("\n"))



from typing import Set, Tuple

import pytest

from json_to_models.generator import MetadataGenerator
from json_to_models.registry import ModelRegistry

# Tuple[Root model name, JSON data] | Set[Tuple[model_name, is_generated]]
test_data = [
    pytest.param(
        ("test_response", {
            "point_south": {
                "x": 0.4,
                "y": 0.2
            },
            "point_north": {
                "x": 0.4,
                "y": 0.2
            },
            "point_west": {
                "x": 0.4,
                "y": 0.2
            },
            "point_east": {
                "x": 0.4,
                "y": 0.2
            }
        }),
        {("TestResponse", False), ("PointEast_PointNorth_PointSouth_PointWest", True)},
        id="merge_models"
    ),
    pytest.param(
        ("results", {
            "points": [
                {
                    "x": 0.4,
                    "y": 0.2
                },
                {
                    "x": 0.4,
                    "y": 0.2
                },
                {
                    "x": 0.4,
                    "y": 0.2
                },
                {
                    "x": 0.4,
                    "y": 0.2
                }
            ],
            "other_points": [
                {
                    "x": 0.4,
                    "y": 0.2
                },
                {
                    "x": 0.4,
                    "y": 0.2
                },
                {
                    "x": 0.4,
                    "y": 0.2
                },
                {
                    "x": 0.4,
                    "y": 0.2
                }
            ]
        }),
        {("Result", False), ("Point", True)},
        id="extract_generic_name"
    ),
    pytest.param(
        (b"Response", {
            "point_south": {
                "x": 0.4,
                "y": 0.2
            }
        }),
        {("Response", False), ("PointSouth", True)},
        id="raw_name_set"
    ),
    pytest.param(
        ("Point", {
            "x": 0.4,
            "y": 0.2,
            "sub": [{
                "x": 0.3,
                "y": 0.7,
                "sub": []
            }]
        }),
        {("Point", False)},
        id="merge_with_name"
    ),
    pytest.param(
        ("Response", {
            "cats": {
                "a": "asdasd"
            },
            "cat": {
                "b": "dfgertdfgdfg"
            }
        }),
        {("Response", False), ("Cat", True), ("Cat_1C", True)},
        id="duplicates"
    ),
]


@pytest.mark.parametrize("value,expected", test_data)
def test_models_names(models_generator: MetadataGenerator, models_registry: ModelRegistry,
                      value: Tuple[str, dict], expected: Set[Tuple[str, bool]]):
    model_name, data = value
    fields = models_generator.generate(data)
    if isinstance(model_name, bytes):
        models_registry.process_meta_data(fields, model_name=model_name.decode())
    elif isinstance(model_name, str):
        ptr = models_registry.process_meta_data(fields)
        ptr.type.name = model_name
    models_registry.merge_models(models_generator)
    models_registry.generate_names()
    names = {(model.name, model.is_name_generated) for model in models_registry.models}
    assert names == expected



import pytest

from json_to_models.dynamic_typing import (
    BooleanString,
    DDict,
    DList,
    DUnion,
    FloatString,
    IntString,
    Null,
    StringLiteral,
    Unknown
)
from json_to_models.generator import MetadataGenerator

# JSON data | MetaData
test_data = [
    pytest.param(1.0, float, id="float"),
    pytest.param(1, int, id="int"),
    pytest.param(True, bool, id="bool"),
    pytest.param("abc", StringLiteral({"abc"}), id="str"),
    pytest.param(None, Null, id="null"),
    pytest.param([], DList(Unknown), id="list_empty"),
    pytest.param([1], DList(int), id="list_single"),
    pytest.param([*range(100)], DList(int), id="list_single_type"),
    pytest.param([1, "a", 2, "c"], DList(DUnion(int, StringLiteral({'a', 'c'}))), id="list_multi"),
    pytest.param("1", IntString, id="int_str"),
    pytest.param("1.0", FloatString, id="float_str"),
    pytest.param("true", BooleanString, id="bool_str"),
    pytest.param({"test_dict_field_a": 1, "test_dict_field_b": "a"}, DDict(DUnion(int, StringLiteral({"a"}))), id="simple_dict"),
    pytest.param({}, DDict(Unknown), id="empty_dict")
]

test_dict = {param.id: param.values[0] for param in test_data}
test_dict_meta = {param.id: param.values[1] for param in test_data}

test_dict_nested = {"b": {"d": test_dict}}
test_dict_nested_meta = {"b": {"d": test_dict_meta}}

test_data += [
    pytest.param(test_dict, test_dict_meta, id="flat_dict"),
    pytest.param(test_dict_nested, test_dict_nested_meta, id="dict"),
]


@pytest.mark.parametrize("value,expected", test_data)
def test_detect_type(models_generator: MetadataGenerator, value, expected):
    assert models_generator._detect_type(value) == expected


def test_convert(models_generator: MetadataGenerator):
    data = {
        "dict_field": {},
        "another_dict_field": {"test_dict_field_a": 1, "test_dict_field_b": "a"},
        "another_dict_field_2": {"test_dict_field_a": 1},
        "another_dict_field_3": {"test_dict_field_a": 1, "test_dict_field_b": 2},
        "int_field": 1,
        "not": False
    }
    meta = models_generator._convert(data)
    assert meta == {
        "dict_field": DDict(Unknown),
        "another_dict_field": DDict(DUnion(int, StringLiteral({"a"}))),
        "another_dict_field_2": DDict(int),
        "another_dict_field_3": DDict(int),
        "int_field": int,
        "not": bool
    }



from builtins import complex

import pytest

from json_to_models.dynamic_typing import DUnion, StringLiteral, get_hash_string

# *args | MetaData
test_dunion = [
    pytest.param(
        [int, int],
        DUnion(int),
        id="unique_types"
    ),
    pytest.param(
        [int, DUnion(int)],
        DUnion(int),
        id="nested_union_&_merge"
    ),
    pytest.param(
        [str, DUnion(int, DUnion(float, complex))],
        DUnion(int, float, complex, str),
        id="complex_merge"
    ),
    pytest.param(
        [str, StringLiteral({'a'})],
        DUnion(str),
        id="str_literal_to_string"
    ),
    pytest.param(
        [StringLiteral({'b'}), StringLiteral({'a'})],
        DUnion(StringLiteral({'a', 'b'})),
        id="str_literal_merge"
    ),
    pytest.param(
        [StringLiteral({str(i)}) for i in range(100)],
        DUnion(str),
        id="str_literal_too_much"
    ),
]


@pytest.mark.parametrize("value,expected", test_dunion)
def test_dunion_creation(value, expected):
    result = DUnion(*value)
    assert result == expected


def test_hash_string():
    a = {'a': int}
    b = {'b': int}
    c = {'a': float}
    assert len(set(map(get_hash_string, (a, b, c)))) == 3

    union = DUnion(str, float)
    h1 = union.to_hash_string()
    union.replace(complex, index=0)
    h2 = union.to_hash_string()
    assert h1 != h2, f"{h1}, {h2}"



from inspect import isclass
from typing import List, Tuple

from .base import GenericModelCodeGenerator, KWAGRS_TEMPLATE, METADATA_FIELD_NAME, sort_kwargs, template
from ..dynamic_typing import (DDict, DList, DOptional, ImportPathList, MetaData, ModelMeta, StringSerializable)

DEFAULT_ORDER = (
    ("default", "default_factory"),
    "*",
    ("metadata",)
)


class DataclassModelCodeGenerator(GenericModelCodeGenerator):
    DC_DECORATOR = template(f"dataclass{{% if kwargs %}}({KWAGRS_TEMPLATE}){{% endif %}}")
    DC_FIELD = template(f"field({KWAGRS_TEMPLATE})")

    def __init__(self, model: ModelMeta, meta=False, dataclass_kwargs: dict = None, **kwargs):
        """
        :param model: ModelMeta instance
        :param meta: Enable generation of metadata as attrib argument
        :param dataclass_kwargs: kwargs for @dataclass() decorators
        :param kwargs:
        """
        super().__init__(model, **kwargs)
        self.no_meta = not meta
        self.dataclass_kwargs = dataclass_kwargs or {}

    @property
    def decorators(self) -> Tuple[ImportPathList, List[str]]:
        imports, decorators = super().decorators
        imports.append(('dataclasses', ['dataclass', 'field']))
        decorators.insert(0, self.DC_DECORATOR.render(kwargs=self.dataclass_kwargs))
        return imports, decorators

    def field_data(self, name: str, meta: MetaData, optional: bool) -> Tuple[ImportPathList, dict]:
        """
        Form field data for template

        :param name: Original field name
        :param meta: Field metadata
        :param optional: Is field optional
        :return: imports, field data
        """
        imports, data = super().field_data(name, meta, optional)
        body_kwargs = {}
        if optional:
            meta: DOptional
            if isinstance(meta.type, DList):
                body_kwargs["default_factory"] = "list"
            elif isinstance(meta.type, DDict):
                body_kwargs["default_factory"] = "dict"
            else:
                body_kwargs["default"] = "None"
                if isclass(meta.type) and issubclass(meta.type, StringSerializable):
                    pass
        elif isclass(meta) and issubclass(meta, StringSerializable):
            pass

        if not self.no_meta and name != data["name"]:
            body_kwargs["metadata"] = {METADATA_FIELD_NAME: name}
        if len(body_kwargs) == 1 and next(iter(body_kwargs.keys())) == "default":
            data["body"] = body_kwargs["default"]
        elif body_kwargs:
            data["body"] = self.DC_FIELD.render(kwargs=sort_kwargs(body_kwargs, DEFAULT_ORDER))
        return imports, data

    @property
    def convert_strings_kwargs(self) -> Tuple[ImportPathList, dict]:
        """
        :return: Imports and Dict with kw-arguments for `json_to_models.models.string_converters.convert_strings` decorator.
        """
        imports, kwargs = super().convert_strings_kwargs
        imports.append(('json_to_models.models', ['ClassType']))
        kwargs["class_type"] = 'ClassType.Dataclass'
        return imports, kwargs



from typing import Dict, List, Optional

import attr

from json_to_models.dynamic_typing import FloatString, IntString
from json_to_models.models import ClassType
from json_to_models.models.string_converters import convert_strings, post_init_converters


def test_post_init_converters():
    from dataclasses import dataclass

    @dataclass
    class A:
        x: IntString
        y: FloatString

        __post_init__ = post_init_converters(['x', 'y'])

    a = A('1', '1.1')
    assert type(a.x) is IntString
    assert type(a.y) is FloatString

    @attr.s
    class A:
        x: IntString = attr.ib()
        y: FloatString = attr.ib()

        __attrs_post_init__ = post_init_converters(['x', 'y'])

    a = A('1', '1.1')
    assert type(a.x) is IntString
    assert type(a.y) is FloatString


def test_convert_strings_decorator():
    from dataclasses import dataclass

    @dataclass
    @convert_strings(['x', 'y'], class_type=ClassType.Dataclass)
    class A:
        x: IntString
        y: FloatString

    @dataclass
    @convert_strings(['x', 'y'], class_type=ClassType.Dataclass)
    class B:
        x: IntString
        y: FloatString

        def __post_init__(self):
            self.x *= 2

    a = A('1', '1.1')
    b = B('1', '1.1')
    assert type(a.x) is IntString
    assert type(a.y) is FloatString
    assert b.x == 2


def test_convert_complex_data():
    from dataclasses import dataclass

    @dataclass
    @convert_strings(['x', 'y#L.S', 'z#D.S', 'a#O.S', 'b#O.L.D.L.S'], class_type=ClassType.Dataclass)
    class A:
        x: IntString
        y: List[IntString]
        z: Dict[str, IntString]
        a: Optional[IntString]
        b: Optional[List[Dict[str, List[IntString]]]]

    a = A('1', '1234', {'s': '2', 'w': '3'}, None,
          [{'a': ['1', '2']}, {'b': ['3', '2']}])

    assert a == A(1, [1, 2, 3, 4], {'s': 2, 'w': 3}, None,
                  [{'a': [1, 2]}, {'b': [3, 2]}])

    @attr.s
    @convert_strings(['x', 'y#L.S', 'z#D.S', 'a#O.S', 'b#O.L.D.L.S'], class_type=ClassType.Attrs)
    class A:
        x: IntString = attr.ib()
        y: List[IntString] = attr.ib()
        z: Dict[str, IntString] = attr.ib()
        a: Optional[IntString] = attr.ib(default=None)
        b: Optional[List[Dict[str, List[IntString]]]] = attr.ib(default=None)

    a = A('1', '1234', {'s': '2', 'w': '3'}, None,
          [{'a': ['1', '2']}, {'b': ['3', '2']}])

    assert a == A(1, [1, 2, 3, 4], {'s': 2, 'w': 3}, None,
                  [{'a': [1, 2]}, {'b': [3, 2]}])



import datetime

import pytest

from json_to_models.dynamic_typing import (BooleanString, FloatString, IntString, IsoDateString, IsoDatetimeString,
                                           IsoTimeString, register_datetime_classes)
from json_to_models.generator import MetadataGenerator

register_datetime_classes()

test_detect_type_data = [
    # Check that string datetime doesn't break default string types
    pytest.param(
        "1",
        IntString,
        id="default_check_int"
    ),
    pytest.param(
        "1.5",
        FloatString,
        id="default_check_float"
    ),
    pytest.param(
        "true",
        BooleanString,
        id="bool"
    ),
    pytest.param(
        "2018-12-31",
        IsoDateString,
        id="date"
    ),
    pytest.param(
        "12:58",
        IsoTimeString,
        id="time"
    ),
    pytest.param(
        "2018-12-31T12:58:12Z",
        IsoDatetimeString,
        id="datetime"
    )
]


@pytest.mark.parametrize("value,expected", test_detect_type_data)
def test_detect_type(models_generator: MetadataGenerator, value, expected):
    result = models_generator._detect_type(value)
    assert result == expected


test_parse_data = [
    pytest.param(
        "true",
        BooleanString(True),
        id="bool"
    ),
    pytest.param(
        "2018-12-31",
        IsoDateString(2018, 12, 31),
        id="date"
    ),
    pytest.param(
        "12:13",
        IsoTimeString(12, 13),
        id="time"
    ),
    pytest.param(
        "04:15:34",
        IsoTimeString(4, 15, 34),
        id="time_seconds"
    ),
    pytest.param(
        "04:15:34.034",
        IsoTimeString(4, 15, 34, 34000),
        id="time_ms"
    ),
    pytest.param(
        "2018-12-04T04:15:34.034000+00:00",
        IsoDatetimeString(2018, 12, 4, 4, 15, 34, 34000, tzinfo=datetime.timezone.utc),
        id="datetime_full"
    ),
    pytest.param(
        "2018-12-04T04:15",
        IsoDatetimeString(2018, 12, 4, 4, 15),
        id="datetime_partial"
    )
]


@pytest.mark.parametrize("value,expected", test_parse_data)
def test_parse(models_generator: MetadataGenerator, value, expected):
    cls = models_generator._detect_type(value)
    result = cls.to_internal_value(value)
    assert result == expected
    assert value in result.to_representation()


def test_replace():
    assert IsoTimeString(14, 12, 57).replace(minute=58, second=32) == IsoTimeString(14, 58, 32)
    assert IsoDateString(2014, 12, 5).replace(day=4, month=5) == IsoDateString(2014, 5, 4)
    assert IsoDatetimeString(2014, 12, 5, 14, 12, 57).replace(minute=58, second=32, day=4, month=5) \
           == IsoDatetimeString(2014, 5, 4, 14, 58, 32)



from collections import OrderedDict

import pytest

from json_to_models.dynamic_typing import ComplexType, DOptional, DTuple, MetaData, ModelMeta, ModelPtr, SingleType
from json_to_models.registry import ModelRegistry

# MetaData | List of models
test_data = [
    pytest.param(
        OrderedDict([
            ('a', int),
            ('b', float)
        ]),
        [
            OrderedDict([
                ('a', int),
                ('b', float)
            ])
        ],
        id="flat"
    ),
    pytest.param(
        OrderedDict([
            ('nested', OrderedDict([
                ('a', int),
                ('b', float)
            ])),
            ('b', float)
        ]),
        [
            OrderedDict([
                ('nested', OrderedDict([
                    ('a', int),
                    ('b', float)
                ])),
                ('b', float)
            ]),
            OrderedDict([
                ('a', int),
                ('b', float)
            ])
        ],
        id="nested"
    ),
    pytest.param(
        OrderedDict([
            ('nested', OrderedDict([
                ('a', int),
                ('b', float)
            ])),
            ('nested_single', DOptional(OrderedDict([
                ('c', int),
                ('d', float)
            ]))),
            ('nested_complex', DTuple(
                OrderedDict([
                    ('nested_2', DOptional(OrderedDict([
                        ('x', int),
                        ('y', float)
                    ]))),
                    ('f', float)
                ]),
                OrderedDict([
                    ('g', int),
                    ('h', float)
                ]),
            ))
        ]),
        [
            # 1A
            OrderedDict([
                ('nested', OrderedDict([
                    ('a', int),
                    ('b', float)
                ])),
                ('nested_single', DOptional(OrderedDict([
                    ('c', int),
                    ('d', float)
                ]))),
                ('nested_complex', DTuple(
                    OrderedDict([
                        ('nested_2', DOptional(OrderedDict([
                            ('x', int),
                            ('y', float)
                        ]))),
                        ('f', float)
                    ]),
                    OrderedDict([
                        ('g', int),
                        ('h', float)
                    ]),
                ))
            ]),
            # 1B
            OrderedDict([
                ('a', int),
                ('b', float)
            ]),
            # 1C
            OrderedDict([
                ('c', int),
                ('d', float)
            ]),
            # 1D
            OrderedDict([
                ('nested_2', DOptional(OrderedDict([
                    ('x', int),
                    ('y', float)
                ]))),
                ('f', float)
            ]),
            # 1E
            OrderedDict([
                ('x', int),
                ('y', float)
            ]),
            # 1F
            OrderedDict([
                ('g', int),
                ('h', float)
            ]),
        ],
        id="complex"
    ),
]

# use it as value in expected model dict to mark this field as cycle reference
cycle_ref = object()


def check_type(meta: MetaData, expected: MetaData):
    if expected is cycle_ref:
        assert isinstance(meta, ModelMeta) or isinstance(meta, ModelPtr)

    elif isinstance(meta, dict):
        assert isinstance(expected, dict)
        for k, v in meta.items():
            check_type(v, expected[k])

    elif isinstance(meta, ModelMeta) or isinstance(meta, ModelPtr):
        check_type(meta.type, expected)

    elif isinstance(meta, SingleType):
        assert isinstance(expected, type(meta))
        check_type(meta.type, expected.type)

    elif isinstance(meta, ComplexType):
        assert isinstance(expected, type(meta))
        for t1, t2 in zip(meta, expected):
            check_type(t1, t2)

    else:
        assert meta == expected


@pytest.mark.parametrize("value,expected", test_data)
def test_registry_process_meta_data(models_registry: ModelRegistry, value, expected):
    models_registry.process_meta_data(value)
    assert len(models_registry.models) == len(expected)
    for model, expected_model in zip(models_registry.models, expected):
        check_type(model, expected_model)


expected_pointers = OrderedDict([
    ("1A", None),
    ("1B", "1A"),
    ("1C", "1A"),
    ("1D", "1A"),
    ("1E", "1D"),
    ("1F", "1A")
])


@pytest.mark.parametrize("value,expected", (pytest.param(
    test_data[2].values[0],
    expected_pointers,
    id="base_test"
),))
def test_registry_pointers(models_registry: ModelRegistry, value, expected):
    models_registry.process_meta_data(value)
    assert len(models_registry.models) == len(expected)
    for model, (index, parent) in zip(models_registry.models, expected.items()):
        assert model.index == index
        ptr = next(iter(model.pointers))
        assert ptr.parent.index if ptr.parent else None == parent



from inspect import isclass
from typing import Any, Dict, Generator, Iterable, List, Tuple, Type, Union

ImportPathList = List[Tuple[str, Union[Iterable[str], str, None]]]


class BaseType:
    def __iter__(self) -> Iterable['MetaData']:
        """
        Yields nested metadata items
        """
        raise NotImplementedError()

    def replace(self, t: Union['MetaData', List['MetaData']], **kwargs) -> 'BaseType':
        """
        Replace nested type in-place

        :param t: Meta type
        :param kwargs: Other args
        :return:
        """
        raise NotImplementedError()

    def to_typing_code(self, types_style: Dict[Union['BaseType', Type['BaseType']], dict]) \
            -> Tuple[ImportPathList, str]:
        """
        Return typing code that represents this metadata and import path of classes that are used in this code

        :param types_style: Hints for .to_typing_code() for different type wrappers
        :return: ((module_name, (class_name, ...)), code)
        """
        raise NotImplementedError()

    @classmethod
    def get_options_for_type(
            cls,
            t: Union['BaseType', Type['BaseType']],
            types_style: Dict[Union['BaseType', Type['BaseType']], dict]
    ) -> dict:
        t_cls = t if isclass(t) else type(t)
        mro = t_cls.__mro__
        for base in mro:
            options = types_style.get(base, ...)
            if options is not Ellipsis:
                return options
        return {}

    def to_hash_string(self) -> str:
        """
        Return unique string that can be used to generate hash of type instance.
        Caches hash value by default. If subclass can mutate (by default it always can)
        then it should define setters to safely invalidate cached value.

        :return: hash string
        """
        # NOTE: Do not override __hash__ function because BaseType instances isn't immutable
        if not getattr(self, '_hash', None):
            self._hash = self._to_hash_string()
        return self._hash

    def _to_hash_string(self) -> str:
        """
        Hash getter method to override

        :return:
        """
        raise NotImplementedError()

    def iter_child(self) -> Generator['MetaData', Any, None]:
        yield self
        for child in self:
            if isinstance(child, BaseType):
                yield from child.iter_child()
            else:
                yield child


class UnknownType(BaseType):
    __slots__ = []

    def __str__(self):
        return "Unknown"

    def __iter__(self) -> Iterable['MetaData']:
        return iter(tuple())

    def replace(self, t: 'MetaData', **kwargs) -> 'UnknownType':
        return self

    def to_typing_code(self, types_style: Dict[Union['BaseType', Type['BaseType']], dict]) \
            -> Tuple[ImportPathList, str]:
        return ([('typing', 'Any')], 'Any')

    def to_hash_string(self) -> str:
        return "Unknown"


class NoneType(BaseType):
    __slots__ = []

    def __str__(self):
        return "NoneType"

    def __iter__(self) -> Iterable['MetaData']:
        return iter(tuple())

    def replace(self, t: 'MetaData', **kwargs) -> 'NoneType':
        return self

    def to_typing_code(self, types_style: Dict[Union['BaseType', Type['BaseType']], dict]) \
            -> Tuple[ImportPathList, str]:
        return ([], 'None')

    def to_hash_string(self) -> str:
        return "NoneType"


Unknown = UnknownType()
Null = NoneType()
MetaData = Union[type, dict, BaseType]


def get_hash_string(t: MetaData):
    if isinstance(t, dict):
        return str(hash(tuple((k, get_hash_string(v)) for k, v in t.items())))
    elif isclass(t):
        return str(t)
    elif isinstance(t, BaseType):
        return t.to_hash_string()



from .base import (
    BaseType, ImportPathList, MetaData, Null, Unknown, get_hash_string
)
from .complex import ComplexType, DDict, DList, DOptional, DTuple, DUnion, SingleType, StringLiteral
from .models_meta import AbsoluteModelRef, ModelMeta, ModelPtr
from .string_datetime import IsoDateString, IsoDatetimeString, IsoTimeString, register_datetime_classes
from .string_serializable import (
    BooleanString, FloatString, IntString, StringSerializable, StringSerializableRegistry, registry
)
from .typing import compile_imports, metadata_to_typing



import argparse
import configparser
import importlib
import itertools
import json
import os.path
import re
import sys
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, Generator, Iterable, List, Tuple, Type, Union

try:
    import ruamel.yaml as yaml
except ImportError:
    try:
        import yaml
    except ImportError:
        yaml = None

from . import __version__ as VERSION
from .dynamic_typing import ModelMeta, register_datetime_classes
from .generator import MetadataGenerator
from .models import ModelsStructureType
from .models.attr import AttrsModelCodeGenerator
from .models.base import GenericModelCodeGenerator, generate_code
from .models.dataclasses import DataclassModelCodeGenerator
from .models.pydantic import PydanticModelCodeGenerator
from .models.structure import compose_models, compose_models_flat
from .registry import (
    ModelCmp, ModelFieldsEquals, ModelFieldsNumberMatch, ModelFieldsPercentMatch, ModelRegistry
)
from .utils import convert_args

STRUCTURE_FN_TYPE = Callable[[Dict[str, ModelMeta]], ModelsStructureType]
bool_js_style = lambda s: {"true": True, "false": False}.get(s, None)


class Cli:
    MODEL_CMP_MAPPING = {
        "percent": convert_args(ModelFieldsPercentMatch, lambda s: float(s) / 100),
        "number": convert_args(ModelFieldsNumberMatch, int),
        "exact": ModelFieldsEquals
    }

    STRUCTURE_FN_MAPPING: Dict[str, STRUCTURE_FN_TYPE] = {
        "nested": compose_models,
        "flat": compose_models_flat
    }

    MODEL_GENERATOR_MAPPING: Dict[str, Type[GenericModelCodeGenerator]] = {
        "base": convert_args(GenericModelCodeGenerator),
        "attrs": convert_args(AttrsModelCodeGenerator, meta=bool_js_style),
        "dataclasses": convert_args(DataclassModelCodeGenerator, meta=bool_js_style,
                                    post_init_converters=bool_js_style),
        "pydantic": convert_args(PydanticModelCodeGenerator),
    }

    def __init__(self):
        self.initialized = False
        self.models_data: Dict[str, Iterable[dict]] = {}  # -m/-l
        self.enable_datetime: bool = False  # --datetime
        self.strings_converters: bool = False  # --strings-converters
        self.max_literals: int = -1  # --max-strings-literals
        self.merge_policy: List[ModelCmp] = []  # --merge
        self.structure_fn: STRUCTURE_FN_TYPE = None  # -s
        self.model_generator: Type[GenericModelCodeGenerator] = None  # -f & --code-generator
        self.model_generator_kwargs: Dict[str, Any] = None

        self.argparser = self._create_argparser()

    def parse_args(self, args: List[str] = None):
        """
        Parse list of command list arguments

        :param args: (Optional) List of arguments
        :return: None
        """
        parser = self.argparser
        namespace = parser.parse_args(args)

        # Extract args
        models: List[Tuple[str, Iterable[Path]]] = [
            (model_name, itertools.chain(*map(_process_path, paths)))
            for model_name, *paths in namespace.model or ()
        ]
        models_lists: List[Tuple[str, Tuple[str, Path]]] = [
            (model_name, (lookup, Path(path)))
            for model_name, lookup, path in namespace.list or ()
        ]
        parser = getattr(FileLoaders, namespace.input_format)
        self.output_file = namespace.output
        self.enable_datetime = namespace.datetime
        disable_unicode_conversion = namespace.disable_unicode_conversion
        self.strings_converters = namespace.strings_converters
        self.max_literals = namespace.max_strings_literals
        merge_policy = [m.split("_") if "_" in m else m for m in namespace.merge]
        structure = namespace.structure
        framework = namespace.framework
        code_generator = namespace.code_generator
        code_generator_kwargs_raw: List[str] = namespace.code_generator_kwargs
        dict_keys_regex: List[str] = namespace.dict_keys_regex
        dict_keys_fields: List[str] = namespace.dict_keys_fields

        self.validate(models, models_lists, merge_policy, framework, code_generator)
        self.setup_models_data(models, models_lists, parser)
        self.set_args(merge_policy, structure, framework, code_generator, code_generator_kwargs_raw,
                      dict_keys_regex, dict_keys_fields, disable_unicode_conversion)

    def run(self):
        if self.enable_datetime:
            register_datetime_classes()
        generator = MetadataGenerator(
            dict_keys_regex=self.dict_keys_regex,
            dict_keys_fields=self.dict_keys_fields
        )
        registry = ModelRegistry(*self.merge_policy)
        for name, data in self.models_data.items():
            meta = generator.generate(*data)
            registry.process_meta_data(meta, name)
        registry.merge_models(generator)
        registry.generate_names()
        structure = self.structure_fn(registry.models_map)
        output = self.version_string + \
                 generate_code(structure, self.model_generator, class_generator_kwargs=self.model_generator_kwargs)
        if self.output_file:
            with open(self.output_file, "w", encoding="utf-8") as f:
                f.write(output)
            return f"Output is written to {self.output_file}"
        else:
            return output

    @property
    def version_string(self):
        return (
            'r"""\n'
            f'generated by json2python-models v{VERSION} at {datetime.now().ctime()}\n'
            f'command: {" ".join(sys.argv)}\n'
            '"""\n'
        )

    def validate(self, models, models_list, merge_policy, framework, code_generator):
        """
        Validate parsed args

        :param models:  List of pairs (model name, list of filesystem path)
        :param models_list: List of pairs (model name, list of lookup expr and filesystem path)
        :param merge_policy: List of merge policies. Each merge policy is either string or string and policy arguments
        :param framework: Framework name (predefined code generator)
        :param code_generator: Code generator import string
        :return:
        """
        names = {name for name, _ in models_list}
        if len(names) != len(models_list):
            raise ValueError("Model names under -l flag should be unique")

        for m in merge_policy:
            if isinstance(m, list):
                if m[0] not in self.MODEL_CMP_MAPPING:
                    raise ValueError(f"Invalid merge policy '{m[0]}', choices are {self.MODEL_CMP_MAPPING.keys()}")
            elif m not in self.MODEL_CMP_MAPPING:
                raise ValueError(f"Invalid merge policy '{m}', choices are {self.MODEL_CMP_MAPPING.keys()}")

        if framework == 'custom' and code_generator is None:
            raise ValueError("You should specify --code-generator to support custom generator")
        elif framework != 'custom' and code_generator is not None:
            raise ValueError("--code-generator argument has no effect without '--framework custom' argument")

    def setup_models_data(
            self,
            models: Iterable[Tuple[str, Iterable[Path]]],
            models_lists: Iterable[Tuple[str, Tuple[str, Path]]],
            parser: 'FileLoaders.T'
    ):
        """
        Initialize lazy loaders for models data
        """
        models_dict: Dict[str, List[Iterable[dict]]] = defaultdict(list)
        for model_name, paths in models:
            models_dict[model_name].append(parser(path) for path in paths)
        for model_name, (lookup, path) in models_lists:
            models_dict[model_name].append(iter_json_file(parser(path), lookup))

        self.models_data = {
            model_name: itertools.chain(*list_of_gen)
            for model_name, list_of_gen in models_dict.items()
        }

    def set_args(
            self,
            merge_policy: List[Union[List[str], str]],
            structure: str,
            framework: str,
            code_generator: str,
            code_generator_kwargs_raw: List[str],
            dict_keys_regex: List[str],
            dict_keys_fields: List[str],
            disable_unicode_conversion: bool
    ):
        """
        Convert CLI args to python representation and set them to appropriate object attributes
        """
        self.merge_policy.clear()
        for merge in merge_policy:
            if isinstance(merge, str):
                name = merge
                args = ()
            else:
                name = merge[0]
                args = merge[1:]
            self.merge_policy.append(self.MODEL_CMP_MAPPING[name](*args))

        self.structure_fn = self.STRUCTURE_FN_MAPPING[structure]

        if framework != "custom":
            self.model_generator = self.MODEL_GENERATOR_MAPPING[framework]
        else:
            module, cls = code_generator.rsplit('.', 1)
            m = importlib.import_module(module)
            self.model_generator = getattr(m, cls)

        self.model_generator_kwargs = dict(
            post_init_converters=self.strings_converters,
            convert_unicode=not disable_unicode_conversion,
            max_literals=self.max_literals
        )
        if code_generator_kwargs_raw:
            for item in code_generator_kwargs_raw:
                if item[0] == '"':
                    item = item[1:]
                if item[-1] == '"':
                    item = item[:-1]
                name, value = item.split("=", 1)
                self.model_generator_kwargs[name] = value

        self.dict_keys_regex = [re.compile(rf"^{r}$") for r in dict_keys_regex] if dict_keys_regex else ()
        self.dict_keys_fields = dict_keys_fields or ()

        self.initialized = True

    @classmethod
    def _create_argparser(cls) -> argparse.ArgumentParser:
        """
        ArgParser factory
        """
        parser = argparse.ArgumentParser(
            formatter_class=argparse.RawTextHelpFormatter,
            description="Convert given json files into Python models."
        )

        parser.add_argument(
            "-m", "--model",
            nargs="+", action="append", metavar=("<Model name>", "<JSON files>"),
            help="Model name and its JSON data as path or unix-like path pattern.\n"
                 "'*',  '**' or '?' patterns symbols are supported.\n\n"
        )
        parser.add_argument(
            "-l", "--list",
            nargs=3, action="append", metavar=("<Model name>", "<JSON key>", "<JSON file>"),
            help="Like -m but given json file should contain list of model data.\n"
                 "If this file contains dict with nested list than you can pass\n"
                 "<JSON key> to lookup. Deep lookups are supported by dot-separated path.\n"
                 "If no lookup needed pass '-' as <JSON key>\n\n"

                 "I.e. for file that contains dict {\"a\": {\"b\": [model_data, ...]}} you should\n"
                 "pass 'a.b' as <JSON key>.\n\n"
        )
        parser.add_argument(
            "-i", "--input-format",
            default="json",
            choices=['json', 'yaml', 'ini'],
            help="Input files parser ('PyYaml' is required to parse yaml files)\n\n"
        )
        parser.add_argument(
            "-o", "--output",
            metavar="FILE", default="",
            help="Path to output file\n\n"
        )
        parser.add_argument(
            "-f", "--framework",
            default="base",
            choices=list(cls.MODEL_GENERATOR_MAPPING.keys()) + ["custom"],
            help="Model framework for which python code is generated.\n"
                 "'base' (default) mean no framework so code will be generated without any decorators\n"
                 "and additional meta-data.\n"
                 "If you pass 'custom' you should specify --code-generator argument\n\n"
        )
        parser.add_argument(
            "-s", "--structure",
            default="flat",
            choices=list(cls.STRUCTURE_FN_MAPPING.keys()),
            help="Models composition style. By default nested models become nested Python classes.\n\n"
        )
        parser.add_argument(
            "--datetime",
            action="store_true",
            help="Enable datetime/date/time strings parsing.\n"
                 "Warn.: This can lead to 6-7 times slowdown on large datasets.\n"
                 "       Be sure that you really need this option.\n\n"
        )
        parser.add_argument(
            "--strings-converters",
            action="store_true",
            help="Enable generation of string types converters (i.e. IsoDatetimeString or BooleanString).\n\n"
        )
        parser.add_argument(
            "--max-strings-literals",
            type=int,
            default=GenericModelCodeGenerator.DEFAULT_MAX_LITERALS,
            metavar='NUMBER',
            help="Generate Literal['foo', 'bar'] when field have less than NUMBER string constants as values.\n"
                 f"Pass 0 to disable. By default NUMBER={GenericModelCodeGenerator.DEFAULT_MAX_LITERALS}"
                 f" (some generator classes could override it)\n\n"
        )
        parser.add_argument(
            "--disable-unicode-conversion", "--no-unidecode",
            action="store_true",
            help="Disabling unicode conversion in fields and class names.\n\n"
        )

        default_percent = f"{ModelFieldsPercentMatch.DEFAULT * 100:.0f}"
        default_number = f"{ModelFieldsNumberMatch.DEFAULT:.0f}"
        parser.add_argument(
            "--merge",
            default=["percent", "number"],
            nargs="+",
            help=(
                f"Merge policy settings. Default is 'percent_{default_percent} number_{default_number}' (percent of field match\n"
                "or number of fields match).\n"
                "Possible values are:\n"
                "'percent[_<percent>]' - two models had a certain percentage of matched field names.\n"
                f"                        Default percent is {default_percent}%%. "
                "Custom value could be i.e. 'percent_95'.\n"
                "'number[_<number>]'   - two models had a certain number of matched field names.\n"
                f"                        Default number of fields is {default_number}.\n"
                "'exact'               - two models should have exact same field names to merge.\n\n"
            )
        )
        parser.add_argument(
            "--dict-keys-regex", "--dkr",
            nargs="+", metavar="RegEx",
            help="List of regular expressions (Python syntax).\n"
                 "If all keys of some dict are match one of them\n"
                 "then this dict will be marked as dict field but not nested model.\n"
                 "Note: ^ and $ tokens will be added automatically but you have to\n"
                 "escape other special characters manually.\n"
        )
        parser.add_argument(
            "--dict-keys-fields", "--dkf",
            nargs="+", metavar="FIELD NAME",
            help="List of model fields names that will be marked as dict fields\n\n"
        )
        parser.add_argument(
            "--code-generator",
            help="Absolute import path to GenericModelCodeGenerator subclass.\n"
                 "Works in pair with '-f custom'\n\n"
        )
        parser.add_argument(
            "--code-generator-kwargs",
            metavar="NAME=VALUE",
            nargs="*", type=str,
            help="List of code generator arguments (for __init__ method).\n"
                 "Each argument should be in following format:\n"
                 "    argument_name=value or \"argument_name=value with space\"\n"
                 "Boolean values should be passed in JS style: true | false"
                 "\n\n"
        )

        return parser


def main():
    import os

    if os.getenv("TRAVIS", None) or os.getenv("FORCE_COVERAGE", None):
        # Enable coverage if it is Travis-CI or env variable FORCE_COVERAGE set to true
        import coverage

        coverage.process_startup()

    cli = Cli()
    cli.parse_args()
    print(cli.run())


def path_split(path: str) -> List[str]:
    """
    Split path into list of components

    :param path: string path
    :return: List of files/patterns
    """
    folders = []
    while True:
        path, folder = os.path.split(path)

        if folder:
            folders.append(folder)
        else:
            if path:
                folders.append(path)
            break
    folders.reverse()
    return folders


class FileLoaders:
    T = Callable[[Path], Union[dict, list]]

    @staticmethod
    def json(path: Path) -> Union[dict, list]:
        with path.open() as fp:
            return json.load(fp)

    @staticmethod
    def yaml(path: Path) -> Union[dict, list]:
        if yaml is None:
            print('Yaml parser is not installed. To parse yaml files PyYaml (or ruamel.yaml) is required.')
            raise ImportError('yaml')
        with path.open() as fp:
            return yaml.safe_load(fp)

    @staticmethod
    def ini(path: Path) -> dict:
        config = configparser.ConfigParser()
        with path.open() as fp:
            config.read_file(fp)
        return {s: dict(config.items(s)) for s in config.sections()}


def dict_lookup(d: Union[dict, list], lookup: str) -> Union[dict, list]:
    """
    Extract nested dictionary value from key path.
    If lookup is "-" returns dict as is.

    :param d: Nested dict
    :param lookup: Dot separated lookup path
    :return: Nested value
    """
    while lookup and lookup != "-":
        split = lookup.split('.', 1)
        if len(split) == 1:
            return d[split[0]]
        key, lookup = split
        d = d[key]
    return d


def iter_json_file(data: Union[dict, list], lookup: str) -> Generator[Union[dict, list], Any, None]:
    """
    Loads given 'path' file, perform lookup and return generator over json list.
    Does not open file until iteration is started.

    :param path: File Path instance
    :param lookup: Dot separated lookup path
    :return:
    """
    l = dict_lookup(data, lookup)
    assert isinstance(l, list), f"Dict lookup return {type(l)} but list is expected, check your lookup path"
    yield from l


def _process_path(path: str) -> Iterable[Path]:
    """
    Convert path pattern into path iterable.
    If non-pattern path is given return tuple of one element: (path,)

    :param path:
    :return:
    """
    split_path = path_split(path)
    clean_path = list(itertools.takewhile(
        lambda part: "*" not in part and "?" not in part,
        split_path
    ))
    pattern_path = split_path[len(clean_path):]

    if clean_path:
        clean_path = os.path.join(*clean_path)
    else:
        clean_path = "."

    if pattern_path:
        pattern_path = os.path.join(*pattern_path)
    else:
        pattern_path = None

    path = Path(clean_path)
    if pattern_path:
        return path.glob(pattern_path)
    else:
        return path,



import imp
import json
import sys
from inspect import isclass

import pydantic
import pytest

from json_to_models.generator import MetadataGenerator
from json_to_models.models.base import generate_code
from json_to_models.models.pydantic import PydanticModelCodeGenerator
from json_to_models.models.structure import compose_models_flat
from json_to_models.registry import ModelRegistry
from .test_script import test_data_path

test_self_validate_pydantic_data = [
    pytest.param(test_data_path / "gists.json", list, id="gists.json"),
    pytest.param(test_data_path / "users.json", list, id="users.json"),
    pytest.param(test_data_path / "unicode.json", dict, id="unicode.json"),
    pytest.param(test_data_path / "photos.json", dict, id="photos.json"),
]


@pytest.mark.parametrize("data,data_type", test_self_validate_pydantic_data)
def test_self_validate_pydantic(data, data_type):
    with data.open() as f:
        data = json.load(f)

    gen = MetadataGenerator(
        dict_keys_fields=['files']
    )
    reg = ModelRegistry()
    if data_type is not list:
        data = [data]
    fields = gen.generate(*data)
    reg.process_meta_data(fields, model_name="TestModel")
    reg.merge_models(generator=gen)
    reg.generate_names()

    structure = compose_models_flat(reg.models_map)
    code = generate_code(structure, PydanticModelCodeGenerator)
    module = imp.new_module("test_models")
    sys.modules["test_models"] = module
    try:
        exec(compile(code, "test_models.py", "exec"), module.__dict__)
    except Exception as e:
        assert not e, code

    import test_models
    for name in dir(test_models):
        cls = getattr(test_models, name)
        if isclass(cls) and issubclass(cls, pydantic.BaseModel):
            cls.update_forward_refs()
    for item in data:
        obj = test_models.TestModel.parse_obj(item)
        assert obj



import copy
import keyword
import re
from typing import Dict, Iterable, List, Tuple, Type, Union

import inflection
from jinja2 import Template
from unidecode import unidecode

from . import INDENT, ModelsStructureType, OBJECTS_DELIMITER
from .string_converters import get_string_field_paths
from .structure import sort_fields
from .utils import indent
from ..dynamic_typing import (AbsoluteModelRef, BaseType, ImportPathList, MetaData,
                              ModelMeta, StringLiteral, compile_imports, metadata_to_typing)
from ..utils import cached_method

METADATA_FIELD_NAME = "J2M_ORIGINAL_FIELD"
KWAGRS_TEMPLATE = "{% for key, value in kwargs.items() %}" \
                  "{{ key }}={{ value }}" \
                  "{% if not loop.last %}, {% endif %}" \
                  "{% endfor %}"

keywords_set = set(keyword.kwlist)
builtins_set = set(__builtins__.keys())
other_common_names_set = {'datetime', 'time', 'date', 'defaultdict', 'schema'}
blacklist_words = frozenset(keywords_set | builtins_set | other_common_names_set)
ones = ['', 'one', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight', 'nine']


def template(pattern: str, indent: str = INDENT) -> Template:
    """
    Remove indent from triple-quotes string and return jinja2.Template instance
    """
    if "\n" in pattern:
        n = len(indent)
        lines = pattern.split("\n")
        for i in (0, -1):
            if not lines[i].strip():
                del lines[i]

        pattern = "\n".join(line[n:] if line[:n] == indent else line
                            for line in lines)
    return Template(pattern)


class GenericModelCodeGenerator:
    """
    Core of model code generator. Extend it to customize fields of model or add some decorators.
    Note that this class has nothing to do with models structure. It only can add nested models as strings.
    """
    BODY = template("""
    {%- for decorator in decorators -%}
    @{{ decorator }}
    {% endfor -%}
    class {{ name }}{% if bases %}({{ bases }}){% endif %}:
    
    {%- for code in nested %}
    {{ code }}
    {% endfor -%}
    
    {%- if fields -%}
    {%- for field in fields %}
        {{ field }}
    {%- endfor %}
    {%- else %}
        pass
    {%- endif -%}
    {%- if extra %}    
    {{ extra }}
    {%- endif -%}
    """)

    STR_CONVERT_DECORATOR = template("convert_strings({{ str_fields }}{%% if kwargs %%}, %s{%% endif %%})"
                                     % KWAGRS_TEMPLATE)
    FIELD: Template = template("{{name}}: {{type}}{% if body %} = {{ body }}{% endif %}")
    DEFAULT_MAX_LITERALS = 10
    default_types_style = {
        StringLiteral: {
            StringLiteral.TypeStyle.use_literals: True
        }
    }

    def __init__(
            self,
            model: ModelMeta,
            max_literals=DEFAULT_MAX_LITERALS,
            post_init_converters=False,
            convert_unicode=True,
            types_style: Dict[Union['BaseType', Type['BaseType']], dict] = None
    ):
        self.model = model
        self.post_init_converters = post_init_converters
        self.convert_unicode = convert_unicode

        resolved_types_style = copy.deepcopy(self.default_types_style)
        types_style = types_style or {}
        for t, style in types_style.items():
            resolved_types_style.setdefault(t, {})
            resolved_types_style[t].update(style)
        resolved_types_style[StringLiteral][StringLiteral.TypeStyle.max_literals] = int(max_literals)
        self.types_style = resolved_types_style

        self.model.set_raw_name(self.convert_class_name(self.model.name), generated=self.model.is_name_generated)

    @cached_method
    def convert_class_name(self, name):
        return prepare_label(name, convert_unicode=self.convert_unicode, to_snake_case=False)

    @cached_method
    def convert_field_name(self, name):
        return prepare_label(name, convert_unicode=self.convert_unicode, to_snake_case=True)

    def generate(self, nested_classes: List[str] = None, bases: str = None, extra: str = "") \
            -> Tuple[ImportPathList, str]:
        """
        :param nested_classes: list of strings that contains classes code
        :return: list of import data, class code
        """
        imports, fields = self.fields
        decorator_imports, decorators = self.decorators
        data = {
            "decorators": decorators,
            "name": self.model.name,
            "bases": bases or [],
            "fields": fields,
            "extra": extra,
        }
        if nested_classes:
            data["nested"] = [indent(s) for s in nested_classes]
        return [*imports, *decorator_imports], self.BODY.render(**data)

    @property
    def decorators(self) -> Tuple[ImportPathList, List[str]]:
        """
        :return: List of imports and List of decorators code (without @)
        """
        imports, decorators = [], []
        if self.post_init_converters:
            str_fields = self.string_field_paths
            decorator_imports, decorator_kwargs = self.convert_strings_kwargs
            if str_fields and decorator_kwargs:
                imports.extend([
                    *decorator_imports,
                    ('json_to_models.models.string_converters', ['convert_strings']),
                ])
                decorators.append(self.STR_CONVERT_DECORATOR.render(str_fields=str_fields, kwargs=decorator_kwargs))
        return imports, decorators

    def field_data(self, name: str, meta: MetaData, optional: bool) -> Tuple[ImportPathList, dict]:
        """
        Form field data for template

        :param name: Original field name
        :param meta: Field metadata
        :param optional: Is field optional
        :return: imports, field data
        """
        imports, typing = metadata_to_typing(meta, types_style=self.types_style)

        data = {
            "name": self.convert_field_name(name),
            "type": typing
        }
        return imports, data

    @property
    def fields(self) -> Tuple[ImportPathList, List[str]]:
        """
        Generate fields strings

        :return: imports, list of fields as string
        """
        required, optional = sort_fields(self.model, unicode_fix=not self.convert_unicode)
        imports: ImportPathList = []
        strings: List[str] = []
        for is_optional, fields in enumerate((required, optional)):
            fields = self._filter_fields(fields)
            for field in fields:
                field_imports, data = self.field_data(field, self.model.type[field], bool(is_optional))
                imports.extend(field_imports)
                strings.append(self.FIELD.render(**data))
        return imports, strings

    def _filter_fields(self, fields):
        return fields

    @property
    def string_field_paths(self) -> List[str]:
        """
        Get paths for convert_strings function
        """
        return [self.convert_field_name(name) + ('#' + '.'.join(path) if path else '')
                for name, path in get_string_field_paths(self.model)]

    @property
    def convert_strings_kwargs(self) -> Tuple[ImportPathList, dict]:
        """
        Override it to enable generation of string types converters

        :return: Imports and Dict with kw-arguments for `json_to_models.models.string_converters.convert_strings` decorator.
        """
        return [], {}


def _generate_code(
        structure: List[dict],
        class_generator: Type[GenericModelCodeGenerator],
        class_generator_kwargs: dict,
        lvl=0
) -> Tuple[ImportPathList, List[str]]:
    """
    Walk thought models structure and covert them into code

    :param structure: Result of compose_models or similar function
    :param class_generator: GenericModelCodeGenerator subclass
    :param class_generator_kwargs: kwags for GenericModelCodeGenerator init
    :param lvl: Recursion depth
    :return: imports, list of first lvl classes
    """
    imports = []
    classes = []
    generators = []
    for data in structure:
        nested_imports, nested_classes = _generate_code(
            data["nested"],
            class_generator,
            class_generator_kwargs,
            lvl=lvl + 1
        )
        imports.extend(nested_imports)
        generators.append((
            class_generator(data["model"], **class_generator_kwargs),
            nested_classes
        ))
    for gen, nested_classes in generators:
        cls_imports, cls_string = gen.generate(nested_classes)
        imports.extend(cls_imports)
        classes.append(cls_string)
    return imports, classes


def generate_code(structure: ModelsStructureType, class_generator: Type[GenericModelCodeGenerator],
                  class_generator_kwargs: dict = None, objects_delimiter: str = OBJECTS_DELIMITER) -> str:
    """
    Generate ready-to-use code

    :param structure: Result of compose_models or similar function
    :param class_generator: GenericModelCodeGenerator subclass
    :param class_generator_kwargs: kwags for GenericModelCodeGenerator init
    :param objects_delimiter: Delimiter between root level classes
    :return: Generated code
    """
    root, mapping = structure
    with AbsoluteModelRef.inject(mapping):
        imports, classes = _generate_code(root, class_generator, class_generator_kwargs or {})
    if imports:
        imports_str = compile_imports(imports) + objects_delimiter
    else:
        imports_str = ""
    return imports_str + objects_delimiter.join(classes) + "\n"


def sort_kwargs(kwargs: dict, ordering: Iterable[Iterable[str]]) -> dict:
    sorted_dict_1 = {}
    sorted_dict_2 = {}
    current = sorted_dict_1
    for group in ordering:
        if isinstance(group, str):
            if group != "*":
                raise ValueError(f"Unknown kwarg group: {group}")
            current = sorted_dict_2
        else:
            for item in group:
                if item in kwargs:
                    value = kwargs.pop(item)
                    current[item] = value
    sorted_dict = {**sorted_dict_1, **kwargs, **sorted_dict_2}
    return sorted_dict


def prepare_label(s: str, convert_unicode: bool, to_snake_case: bool) -> str:
    if convert_unicode:
        s = unidecode(s)
    s = re.sub(r"\W", "", s)
    if not ('a' <= s[0].lower() <= 'z'):
        if '0' <= s[0] <= '9':
            s = ones[int(s[0])] + "_" + s[1:]
    if to_snake_case:
        s = inflection.underscore(s)
    if s in blacklist_words:
        s += "_"
    return s



from functools import wraps
from inspect import isclass
from typing import Any, Callable, List, Optional, Tuple

from . import ClassType
from ..dynamic_typing import (
    BaseType,
    DDict,
    DList,
    DOptional,
    DUnion,
    MetaData,
    ModelMeta,
    ModelPtr,
    StringLiteral,
    StringSerializable
)
from ..dynamic_typing.base import NoneType


def convert_strings(str_field_paths: List[str], class_type: Optional[ClassType] = None,
                    method: Optional[str] = None) -> Callable[[type], type]:
    """
    Decorator factory. Set up post-init method to convert strings fields values into StringSerializable types

    If field contains complex data type path should be consist of field name and dotted list of tokens:

    * `S` - string component
    * `O` - Optional
    * `L` - List
    * `D` - Dict

    So if field `'bar'` has type `Optional[List[List[IntString]]]` field path would be `'bar#O.L.L.S'`

    ! If type is too complex i.e. Union[List[IntString], List[List[IntString]]]
    you can't specify field path and such field would be ignored

    To specify name of post-init method you should provide it by class_type argument or directly by method argument:

    >>> convert_strings([...], class_type=ClassType.Attrs)

    is equivalent of

    >>> convert_strings([...], method="__attrs_post_init__")

    :param str_field_paths: Paths of StringSerializable fields (field name or field name + typing path)
    :param class_type: attrs | dataclass - type of decorated class
    :param method: post-init method name
    :return: Class decorator
    """
    method = {
        ClassType.Attrs: '__attrs_post_init__',
        ClassType.Dataclass: '__post_init__',
        None: method
    }.get(class_type)

    def decorator(cls: type) -> type:
        if hasattr(cls, method):
            old_fn = getattr(cls, method)

            @wraps(old_fn)
            def __post_init__(self, *args, **kwargs):
                post_init_converters(str_field_paths)(self)
                old_fn(self, *args, **kwargs)

            setattr(cls, method, __post_init__)
        else:
            fn = post_init_converters(str_field_paths)
            fn.__name__ = method
            setattr(cls, method, fn)

        return cls

    return decorator


def post_init_converters(str_fields: List[str], wrap_fn=None):
    """
    Method factory. Return post_init method to convert string into StringSerializable types
    To override generated __post_init__ you can call it directly:

    >>> def __post_init__(self):
    ...     post_init_converters(['a', 'b'])(self)

    :param str_fields: names of StringSerializable fields
    :return: __post_init__ method
    """

    def __post_init__(self):
        # `S` - string component
        # `O` - Optional
        # `L` - List
        # `D` - Dict
        for name in str_fields:
            if '#' in name:
                name, path_str = name.split('#')
                path: List[str] = path_str.split('.')
            else:
                path = ['S']

            new_value = _process_string_field_value(
                path=path,
                value=getattr(self, name),
                current_type=self.__annotations__[name]
            )
            setattr(self, name, new_value)

    if wrap_fn:
        __post_init__ = wraps(wrap_fn)(__post_init__)

    return __post_init__


def _process_string_field_value(path: List[str], value: Any, current_type: Any, optional=False) -> Any:
    token, *path = path
    if token == 'S':
        try:
            value = current_type.to_internal_value(value)
        except (ValueError, TypeError) as e:
            if not optional:
                raise e
        return value
    elif token == 'O':
        return _process_string_field_value(
            path=path,
            value=value,
            current_type=current_type.__args__[0],
            optional=True
        )
    elif token == 'L':
        t = current_type.__args__[0]
        return [
            _process_string_field_value(path, item, current_type=t, optional=optional)
            for item in value
        ]
    elif token == 'D':
        t = current_type.__args__[1]
        return {
            key: _process_string_field_value(path, item, current_type=t, optional=optional)
            for key, item in value.items()
        }
    else:
        raise ValueError(f"Unknown token {token}")


def get_string_field_paths(model: ModelMeta) -> List[Tuple[str, List[str]]]:
    """
    Return paths for convert_strings function of given model

    :return: Paths with raw names
    """
    # `S` - string component
    # `O` - Optional
    # `L` - List
    # `D` - Dict
    str_fields: List[Tuple[str, List[str]]] = []
    for name, t in model.type.items():

        # Walk through nested types
        paths: List[List[str]] = []
        tokens: List[Tuple[MetaData, List[str]]] = [(t, ['#'])]
        while tokens:
            tmp_type, path = tokens.pop()
            if isclass(tmp_type):
                if issubclass(tmp_type, StringSerializable):
                    paths.append(path + ['S'])
            elif isinstance(tmp_type, BaseType):
                cls = type(tmp_type)
                if cls is DOptional:
                    token = 'O'
                elif cls is DList:
                    token = 'L'
                elif cls is DDict:
                    token = 'D'
                elif cls in (DUnion, ModelPtr):
                    # We could not resolve Union
                    paths = []
                    break
                elif cls is NoneType:
                    continue
                elif cls in (StringLiteral,):
                    continue
                else:
                    raise TypeError(f"Unsupported meta-type for converter path {cls}")

                for nested_type in tmp_type:
                    tokens.append((nested_type, path + [token]))
        paths: List[str] = ["".join(p[1:]) for p in paths]
        if len(paths) != 1:
            continue

        path = paths.pop()
        if path == 'S':
            str_fields.append((name, []))
        else:
            str_fields.append((name, path))

    return str_fields



from collections import OrderedDict
from random import shuffle

import pytest

from json_to_models.dynamic_typing import DOptional, DUnion, FloatString, IntString
from json_to_models.generator import MetadataGenerator

# List of fields sets | result field set
test_data = [
    pytest.param(
        [{'a': int}, {'a': int, 'b': int}],
        {'a': int, 'b': DOptional(int)},
        id="optional_fields"
    ),
    pytest.param(
        [{'a': int}, {'a': float}],
        {'a': DUnion(int, float)},
        id="union"
    ),
    pytest.param(
        [{'a': DUnion(str, bool)}, {'a': float}],
        {'a': DUnion(str, bool, float)},
        id="merge_unions"
    ),
    pytest.param(
        [{'a': DUnion(str, bool)}, {'a': DUnion(int, float)}],
        {'a': DUnion(str, bool, int, float)},
        id="merge_unions2"
    ),
    pytest.param(
        [{'a': int}, {}, {'a': int}, {}],
        {'a': DOptional(int)},
        id="merge_optional_int"
    ),
    pytest.param(
        [{'a': str}, {}, {'a': int}, {}],
        {'a': DOptional(DUnion(str, int))},
        id="merge_optional_str_int"
    ),
    pytest.param(
        [{'a': DUnion(int)}, {}, {'a': int}, {}],
        {'a': DOptional(int)},
        id="merge_optional_single_union"
    ),
    pytest.param(
        [{'a': DUnion(int)}, {'a': int}],
        {'a': int},
        id="merge_single_union"
    ),
    pytest.param(
        [
            {'a': DUnion(int, str)},
            {'a': DUnion(int, FloatString)},
            {'a': DUnion(int, IntString)},
            {'a': DUnion(int, str)},
            {'a': DUnion(int, IntString)},
            {'a': DUnion(int, str)},
            {'a': DUnion(int, FloatString)},
            {'a': DUnion(int, IntString)},
            {'a': DUnion(int, str)},
            {'a': DUnion(int, IntString)},
        ],
        {'a': DUnion(int, str, FloatString, IntString)},
        id="merge_unions_of_pseudo_strings"
    ),
    pytest.param(
        [
            {'a': int},
            {},
            {'a': DOptional(int)},
        ],
        {'a': DOptional(int)},
        id="merge_optionals_and_nulls"
    ),
    # This functional is moved to _optimize_type
    # pytest.param(
    #     [{'d': {'x': int}}, {'d': {'x': NoneType}}],
    #     {'d': {'x': DUnion(int, NoneType)}},
    #     id="merge_nested"
    # )
]


@pytest.mark.parametrize("value,expected", test_data)
def test_merge_field_sets(models_generator: MetadataGenerator, value, expected):
    shuffled = value[:]
    shuffle(shuffled)
    for v in (value, reversed(value), shuffled):
        result = models_generator.merge_field_sets(v)
        if isinstance(result, OrderedDict):
            result = dict(result)
        assert result == expected



import imp
import json
import subprocess
import sys
import tempfile
from pathlib import Path
from time import time
from typing import Tuple

import pytest

tmp_dir = tempfile.TemporaryDirectory(f"-pytest-{time()}")
tmp_path = Path(tmp_dir.name)
test_data_path = (Path(__file__) / ".." / "data").resolve().absolute()
if not test_data_path.exists():
    test_data_path = Path("./test/test_cli/data").resolve().absolute()


# Create fixture to auto cleanup tmp directory after tests
@pytest.fixture(scope="session", autouse=True)
def tmp_dir_cleanup():
    yield tmp_dir
    tmp_dir.cleanup()


# download GitHub Gist dataset into tmp folder
with (test_data_path / "gists.json").open("r") as f:
    gists = json.load(f)
assert type(gists) is list and gists and type(gists[0]) is dict

for item in gists:
    with (tmp_path / f"{item['id']}.gist").open("w") as f:
        json.dump(item, f)

# detect script path
setuptools_script = subprocess.call(["json2models"], shell=True) == 0
if setuptools_script:
    executable = "json2models"
else:
    python_path = sys.executable.replace('\\', '/')
    executable = f"{python_path} -m json_to_models"


def test_help():
    c = f"{executable} -h"
    proc = subprocess.Popen(c, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    assert not stderr, stderr
    assert stdout, stdout
    assert proc.returncode == 0
    print(stdout.decode())


test_commands = [
    pytest.param(f"""{executable} -l Photo items "{test_data_path / 'photos.json'}" """, id="list1"),
    pytest.param(f"""{executable} -l User - "{test_data_path / 'users.json'}" """, id="list2"),
    pytest.param(f"""{executable} -m Photos "{test_data_path / 'photos.json'}" """, id="model1"),

    pytest.param(f"""{executable} -l Photo items "{test_data_path / 'photos.json'}" \
                                  -m Photos "{test_data_path / 'photos.json'}" """,
                 id="list1_model1"),

    pytest.param(f"""{executable} -l Photo items "{test_data_path / 'photos.json'}" \
                                  -l User - "{test_data_path / 'users.json'}" """,
                 id="list1_list2"),

    pytest.param(f"""{executable} -m Gist "{tmp_path / '*.gist'}" --dkf files""",
                 id="gists"),
    pytest.param(f"""{executable} -m Gist "{tmp_path / '*.gist'}" --dkf files --datetime""",
                 id="gists_datetime"),
    pytest.param(f"""{executable} -m Gist "{tmp_path / '*.gist'}" --dkf files --merge percent number_10""",
                 id="gists_merge_policy"),
    pytest.param(f"""{executable} -m Gist "{tmp_path / '*.gist'}" --dkf files --merge exact""",
                 id="gists_no_merge"),
    pytest.param(f"""{executable} -m Gist "{tmp_path / '*.gist'}" --dkf files --datetime --strings-converters""",
                 id="gists_strings_converters"),

    pytest.param(f"""{executable} -l User - "{test_data_path / 'users.json'}" --strings-converters""",
                 id="users_strings_converters"),
    pytest.param(f"""{executable} -m SomeUnicode "{test_data_path / 'unicode.json'}" """,
                 id="convert_unicode"),
    pytest.param(f"""{executable} -m SomeUnicode "{test_data_path / 'unicode.json'}" --no-unidecode""",
                 id="dont_convert_unicode"),
    pytest.param(f"""{executable} -m SomeUnicode "{test_data_path / 'unicode.json'}" --disable-unicode-conversion""",
                 id="dont_convert_unicode_2"),
    pytest.param(f"""{executable} -m YamlFile "{test_data_path / 'spotify-swagger.yaml'}" -i yaml""",
                 id="yaml_file"),
    pytest.param(f"""{executable} -m IniFile "{test_data_path / 'file.ini'}" -i ini""",
                 id="ini_file"),
]


def _validate_result(proc: subprocess.Popen, output=None, output_file: Path = None) -> Tuple[str, str]:
    stdout, stderr = map(bytes.decode, proc.communicate())
    if output_file:
        assert output is None
        with output_file.open(encoding='utf-8') as f:
            output = f.read()
    if output:
        stdout = output
    assert not stderr, stderr
    assert stdout, stdout
    assert proc.returncode == 0
    # Note: imp package is deprecated but I can't find a way to create dummy module using importlib
    module = imp.new_module("test_model")
    sys.modules["test_model"] = module
    try:
        exec(compile(stdout, "test_model.py", "exec"), module.__dict__)
    except Exception as e:
        assert not e, stdout
    return stdout, stderr


@pytest.mark.parametrize("command", test_commands)
def test_script(command):
    proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = _validate_result(proc)
    print(stdout)


@pytest.mark.parametrize("command", test_commands)
def test_script_flat(command):
    command += " -s flat"
    proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = _validate_result(proc)
    print(stdout)


@pytest.mark.parametrize("command", test_commands)
def test_script_attrs(command):
    command += " -f attrs"
    proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = _validate_result(proc)
    assert "@attr.s" in stdout
    print(stdout)


@pytest.mark.parametrize("command", test_commands)
def test_script_pydantic(command):
    command += " -f pydantic"
    # Pydantic has native (str) -> (builtin_type) converters
    command = command.replace('--strings-converters', '')
    proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = _validate_result(proc)
    assert "(BaseModel):" in stdout
    print(stdout)


@pytest.mark.parametrize("command", test_commands)
def test_script_pydantic_disable_literals(command):
    command += " -f pydantic --code-generator-kwargs max_literals=0"
    # Pydantic has native (str) -> (builtin_type) converters
    command = command.replace('--strings-converters', '')
    proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = _validate_result(proc)
    assert "(BaseModel):" in stdout
    assert "Literal" not in stdout
    print(stdout)


@pytest.mark.parametrize("command", test_commands)
def test_script_dataclasses(command):
    command += " -f dataclasses"
    proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = _validate_result(proc)
    assert "@dataclass" in stdout
    print(stdout)


@pytest.mark.parametrize("command", test_commands)
def test_script_custom(command):
    command += " -f custom --code-generator json_to_models.models.attr.AttrsModelCodeGenerator"
    command += ' --code-generator-kwargs "meta=true"'
    proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = _validate_result(proc)
    assert "@attr.s" in stdout
    print(stdout)


wrong_arguments_commands = [
    pytest.param(f"""{executable} -l Model items "{test_data_path / 'photos.json'}" \
                                  -l Model - "{test_data_path / 'users.json'}" """, id="duplicate_name"),
    pytest.param(f"""{executable} -l Model items "{test_data_path / 'photos.json'}" --merge unknown""",
                 id="wrong_merge_policy"),
    pytest.param(f"""{executable} -l Model items "{test_data_path / 'photos.json'}" --merge unknown_10""",
                 id="wrong_merge_policy"),
    pytest.param(f"""{executable} -l Model items "{test_data_path / 'photos.json'}" -f custom""",
                 id="custom_model_generator_without_class_link"),
    pytest.param(f"""{executable} -l Model items "{test_data_path / 'photos.json'}" --code-generator test""",
                 id="class_link_without_custom_model_generator_enabled"),
]


@pytest.mark.xfail
@pytest.mark.parametrize("command", wrong_arguments_commands)
def test_wrong_arguments(command):
    print("Command:", command)
    proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    _validate_result(proc)


@pytest.mark.parametrize("command", test_commands)
def test_script_output_file(command):
    file = tmp_path / 'out.py'
    command += f" -o {file}"
    proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = _validate_result(proc, output_file=file)
    print(stdout)



import threading
from typing import Dict, List, Optional, Set, Tuple, Type, Union

import inflection

from . import BaseType
from .base import ImportPathList, MetaData
from .complex import SingleType
from ..utils import distinct_words


class ModelMeta(SingleType):
    WORDS_SEPARATOR = "_"

    def __init__(self, t: MetaData, index, _original_fields=None):
        super().__init__(t)
        self.original_fields: List[List[str]] = _original_fields or [list(self.type.keys())]
        self.index: str = index
        self.pointers: Set[ModelPtr] = set()
        self.child_pointers: Set[ModelPtr] = set()  # parent ref (pointers that have ptr.parent == self)
        self._name: Optional[str] = None
        self._name_generated: Optional[bool] = None

    def __str__(self):
        return f"Model#{self.index}" + ("-" + self._name if self._name else "")

    def __repr__(self):
        return f"<{self}>"

    def __eq__(self, other):
        if isinstance(other, dict):
            return self.type == other
        else:
            return super().__eq__(other)

    def __hash__(self):
        return hash(self.index)

    def generate_name(self):
        """
        Generate model name based on fields to which his model is assigned.
        Will overwrite existed name so check is_name_generated before call this method
        """
        base_names = (inflection.singularize(inflection.underscore(ptr.parent_field_name))
                      for ptr in self.pointers if ptr.parent is not None)
        filtered_names = distinct_words(*base_names)
        new_name = self.name_joiner(*map(inflection.camelize, sorted(filtered_names)))
        if new_name:
            self._name = new_name
            self._name_generated = True

    @classmethod
    def name_joiner(cls, *names: str) -> str:
        """
        Join words to form a model name. Uses in different places so override WORDS_SEPARATOR to change it globally.
        """
        return cls.WORDS_SEPARATOR.join(names)

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str):
        """
        Convert given name to singular form and CamelCase format
        """
        self._name = inflection.camelize(inflection.singularize(name))
        self._name_generated = False

    @name.deleter
    def name(self):
        self._name = None

    def set_raw_name(self, name, generated=False):
        """
        Set model name and is_name_generated flag as is without any conversion made
        """
        self._name = name
        self._name_generated = generated

    @property
    def is_name_generated(self):
        return self._name_generated

    def connect(self, ptr: 'ModelPtr'):
        self.pointers.add(ptr)

    def disconnect(self, ptr: 'ModelPtr'):
        self.pointers.remove(ptr)

    def add_child_ref(self, ptr: 'ModelPtr'):
        self.child_pointers.add(ptr)

    def remove_child_ref(self, ptr: 'ModelPtr'):
        self.child_pointers.remove(ptr)

    def to_typing_code(self, types_style: Dict[Union[BaseType, Type[BaseType]], dict]) \
            -> Tuple[ImportPathList, str]:
        if self.name is None:
            raise ValueError('Model without name can not be typed')
        return [], self.name


class ModelPtr(SingleType):
    """
    Model wrapper (pointer)
    """
    type: ModelMeta

    def __init__(self, meta: ModelMeta, parent: ModelMeta = None, parent_field_name: str = None):
        super().__init__(meta)
        self.parent = parent
        self.parent_field_name = parent_field_name
        meta.connect(self)
        if parent:
            parent.add_child_ref(self)

    def __hash__(self):
        return id(self)

    def replace(self, t: ModelMeta, **kwargs) -> 'ModelPtr':
        self.type.disconnect(self)
        super().replace(t, **kwargs)
        self.type.connect(self)
        return self

    def replace_parent(self, t: ModelMeta, **kwargs) -> 'ModelPtr':
        self._hash = None
        self.parent.remove_child_ref(self)
        self.parent = t
        self.parent.add_child_ref(self)
        return self

    def to_typing_code(self, types_style: Dict[Union[BaseType, Type[BaseType]], dict]) \
            -> Tuple[ImportPathList, str]:
        return AbsoluteModelRef(self.type).to_typing_code(types_style)

    def _to_hash_string(self) -> str:
        return f"{type(self).__name__}_#{self.type.index}"


ContextInjectionType = Dict[ModelMeta, Union[ModelMeta, str]]


class AbsoluteModelRef:
    """
    Model forward absolute references. Using ContextManager to inject real models paths into typing code.
    Forward reference is the typing string like ``List['MyModel']``.
    If the model is defined as child model and is used by another nested model
    than the reference to this model should be an absolute path:

    class Model:
        class GenericChildModel:
            ...

        class NestedModel:
            data: 'Model.GenericChildModel'  # <--- this

    This information is only available at the models code generation stage
    while typing code is generated from raw metadata and passing this absolute path as argument
    to each ModelPtr would be annoying.

    Usage:

    with AbsoluteModelRef.inject({TestModel: "ParentModelName"}):
        <some code generation>
    """

    class Context:
        data = threading.local()
        data.context: ContextInjectionType = None

        def __init__(self, patches: ContextInjectionType):
            self.context: ContextInjectionType = patches
            self._old: ContextInjectionType = None

        def __enter__(self):
            self._old = self.data.context
            self.data.context = self.context

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.data.context = self._old

    @classmethod
    def inject(cls, patches: ContextInjectionType):
        context = cls.Context(patches)
        return context

    def __init__(self, model: ModelMeta):
        self.model = model

    def to_typing_code(self, types_style: Dict[Union[BaseType, Type[BaseType]], dict]) \
            -> Tuple[ImportPathList, str]:
        context_data = self.Context.data.context
        if context_data:
            model_path = context_data.get(self.model, "")
            if isinstance(model_path, ModelMeta):
                model_path = model_path.name
        else:
            model_path = ""
        imports, model = self.model.to_typing_code(types_style)
        s = ".".join(filter(None, (model_path, model)))
        return imports, f"'{s}'"



import operator
from datetime import date, datetime, time
from typing import Any, Optional, Type, Union

import dateutil.parser

from .string_serializable import StringSerializable, StringSerializableRegistry, registry

_dt_args_getter = operator.attrgetter('year', 'month', 'day', 'hour', 'minute', 'second', 'microsecond', 'tzinfo')
_d_args_getter = operator.attrgetter('year', 'month', 'day')
_t_args_getter = operator.attrgetter('hour', 'minute', 'second', 'microsecond', 'tzinfo')


def extend_datetime(d: Union[date, time, datetime], cls: Union[Type[date], Type[time], Type[datetime]]) -> Any:
    """
    Wrap datetime object into datetime subclass

    :param d: date/time/datetime instance
    :param cls: datetime subclass
    :return:
    """
    if isinstance(d, datetime):
        args = _dt_args_getter
    elif isinstance(d, time):
        args = _t_args_getter
    else:
        args = _d_args_getter
    return cls(*args(d))


_check_values_date = (
    datetime(2018, 1, 2, 0, 4, 5, 678, tzinfo=None),
    datetime(2018, 1, 2, 9, 4, 5, 678, tzinfo=None)
)


def is_date(s: str) -> Optional[date]:
    """
    Return date instance if given string is a date and None otherwise

    :param s: string
    :return: date or None
    """
    # dateutil.parser.parse replaces missing parts of datetime with values from default value
    # so if there is hour part in given string then d1 and d2 would be equal and string is not pure date
    d1 = dateutil.parser.parse(s, default=_check_values_date[0])
    d2 = dateutil.parser.parse(s, default=_check_values_date[1])
    return None if d1 == d2 else d1.date()


_check_values_time = (
    datetime(2018, 10, 11),
    datetime(2018, 12, 30)
)


def is_time(s: str) -> Optional[time]:
    """
    Return time instance if given string is a time and None otherwise

    :param s: string
    :return: time or None
    """
    d1 = dateutil.parser.parse(s, default=_check_values_time[0])
    d2 = dateutil.parser.parse(s, default=_check_values_time[1])
    return None if d1 == d2 else d1.time()


class IsoDateString(StringSerializable, date):
    """
    Parse date using dateutil.parser.isoparse. Representation format always is ``YYYY-MM-DD``.
    You can override to_representation method to customize it. Just don't forget to call registry.remove(IsoDateString)
    """
    actual_type = date

    @classmethod
    def to_internal_value(cls, value: str) -> 'IsoDateString':
        if not is_date(value):
            raise ValueError(f"'{value}' is not valid date")
        dt = dateutil.parser.isoparse(value)
        return extend_datetime(dt.date(), cls)

    def to_representation(self):
        return self.isoformat()

    def replace(self, *args, **kwargs) -> 'IsoDateString':
        # noinspection PyTypeChecker
        return date.replace(self, *args, **kwargs)


class IsoTimeString(StringSerializable, time):
    """
    Parse time using dateutil.parser.parse. Representation format always is ``hh:mm:ss.ms``.
    You can override to_representation method to customize it.
    """
    actual_type = time

    @classmethod
    def to_internal_value(cls, value: str) -> 'IsoTimeString':
        t = is_time(value)
        if not t:
            raise ValueError(f"'{value}' is not valid time")
        return extend_datetime(t, cls)

    def to_representation(self):
        return self.isoformat()

    def replace(self, *args, **kwargs) -> 'IsoTimeString':
        # noinspection PyTypeChecker
        return time.replace(self, *args, **kwargs)


class IsoDatetimeString(StringSerializable, datetime):
    """
    Parse datetime using dateutil.parser.isoparse.
    Representation format always is ``YYYY-MM-DDThh:mm:ss.ms`` (datetime.isoformat method).
    """
    actual_type = datetime

    @classmethod
    def to_internal_value(cls, value: str) -> 'IsoDatetimeString':
        dt = dateutil.parser.isoparse(value)
        return extend_datetime(dt, cls)

    def to_representation(self):
        return self.isoformat()

    def replace(self, *args, **kwargs) -> 'IsoDatetimeString':
        # noinspection PyTypeChecker
        return datetime.replace(self, *args, **kwargs)


def register_datetime_classes(registry: StringSerializableRegistry = registry):
    """
    Register datetime classes in given registry (using default registry if no arguments is passed).
    Date parsing is expensive operation so this classes are disabled by default
    """
    registry.add(cls=IsoDateString)
    registry.add(cls=IsoTimeString)
    registry.add(cls=IsoDatetimeString)



from typing import Dict, List

import pytest

from json_to_models.dynamic_typing import (
    DDict,
    DList,
    DOptional,
    DUnion,
    FloatString,
    IntString,
    ModelMeta,
    compile_imports,
)
from json_to_models.models.base import generate_code
from json_to_models.models.pydantic import PydanticModelCodeGenerator
from json_to_models.models.structure import sort_fields
from test.test_code_generation.test_models_code_generator import model_factory, trim

# Data structure:
# pytest.param id -> {
#   "model" -> (model_name, model_metadata),
#   test_name -> expected, ...
# }
test_data = {
    "base": {
        "model": ("Test", {
            "foo": int,
            "Bar": int,
            "baz": float
        }),
        "fields_data": {
            "foo": {
                "name": "foo",
                "type": "int"
            },
            "Bar": {
                "name": "bar",
                "type": "int",
                "body": 'Field(..., alias="Bar")'
            },
            "baz": {
                "name": "baz",
                "type": "float"
            }
        },
        "fields": {
            "imports": "",
            "fields": [
                f"foo: int",
                f'bar: int = Field(..., alias="Bar")',
                f"baz: float",
            ]
        },
        "generated": trim(f"""
        from pydantic import BaseModel, Field
        
        
        class Test(BaseModel):
            foo: int
            bar: int = Field(..., alias="Bar")
            baz: float
        """)
    },
    "complex": {
        "model": ("Test", {
            "foo": int,
            "baz": DOptional(DList(DList(str))),
            "bar": DOptional(IntString),
            "qwerty": FloatString,
            "asdfg": DOptional(int),
            "dict": DDict(int),
            "not": bool,
            "1day": int,
            "день_недели": str,
        }),
        "fields_data": {
            "foo": {
                "name": "foo",
                "type": "int"
            },
            "baz": {
                "name": "baz",
                "type": "Optional[List[List[str]]]",
                "body": "[]"
            },
            "bar": {
                "name": "bar",
                "type": "Optional[int]",
                "body": "None"
            },
            "qwerty": {
                "name": "qwerty",
                "type": "float"
            },
            "asdfg": {
                "name": "asdfg",
                "type": "Optional[int]",
                "body": "None"
            },
            "dict": {
                "name": "dict_",
                "type": "Dict[str, int]",
                "body": 'Field(..., alias="dict")'
            },
            "not": {
                "name": "not_",
                "type": "bool",
                "body": 'Field(..., alias="not")'
            },
            "1day": {
                "name": "one_day",
                "type": "int",
                "body": 'Field(..., alias="1day")'
            },
            "день_недели": {
                "name": "den_nedeli",
                "type": "str",
                "body": 'Field(..., alias="день_недели")'
            }
        },
        "generated": trim(f"""
        from pydantic import BaseModel, Field
        from typing import Dict, List, Optional


        class Test(BaseModel):
            foo: int
            qwerty: float
            dict_: Dict[str, int] = Field(..., alias="dict")
            not_: bool = Field(..., alias="not")
            one_day: int = Field(..., alias="1day")
            den_nedeli: str = Field(..., alias="день_недели")
            baz: Optional[List[List[str]]] = []
            bar: Optional[int] = None
            asdfg: Optional[int] = None
        """)
    },
    "converters": {
        "model": ("Test", {
            "a": int,
            "b": IntString,
            "c": DOptional(FloatString),
            "d": DList(DList(DList(IntString))),
            "e": DDict(IntString),
            "u": DUnion(DDict(IntString), DList(DList(IntString))),
        }),
        "generated": trim("""
        from pydantic import BaseModel, Field
        from typing import Dict, List, Optional, Union


        class Test(BaseModel):
            a: int
            b: int
            d: List[List[List[int]]]
            e: Dict[str, int]
            u: Union[Dict[str, int], List[List[int]]]
            c: Optional[float] = None
        """)
    }
}

test_data_unzip = {
    test: [
        pytest.param(
            model_factory(*data["model"]),
            data[test],
            id=id
        )
        for id, data in test_data.items()
        if test in data
    ]
    for test in ("fields_data", "fields", "generated")
}


@pytest.mark.parametrize("value,expected", test_data_unzip["fields_data"])
def test_fields_data_attr(value: ModelMeta, expected: Dict[str, dict]):
    gen = PydanticModelCodeGenerator(value)
    required, optional = sort_fields(value)
    for is_optional, fields in enumerate((required, optional)):
        for field in fields:
            field_imports, data = gen.field_data(field, value.type[field], bool(is_optional))
            assert data == expected[field]


@pytest.mark.parametrize("value,expected", test_data_unzip["fields"])
def test_fields_attr(value: ModelMeta, expected: dict):
    expected_imports: str = expected["imports"]
    expected_fields: List[str] = expected["fields"]
    gen = PydanticModelCodeGenerator(value)
    imports, fields = gen.fields
    imports = compile_imports(imports)
    assert imports == expected_imports
    assert fields == expected_fields


@pytest.mark.parametrize("value,expected", test_data_unzip["generated"])
def test_generated_attr(value: ModelMeta, expected: str):
    generated = generate_code(
        (
            [{"model": value, "nested": []}],
            {}
        ),
        PydanticModelCodeGenerator,
        class_generator_kwargs={}
    )
    assert generated.rstrip() == expected, generated



import json
import sys
from pathlib import Path

import pytest

from json_to_models.cli import _process_path, dict_lookup, iter_json_file, path_split
from json_to_models.utils import convert_args

echo = lambda *args, **kwargs: (args, kwargs)
test_dict = {"user":
    {
        "id": 1,
        "name": "Leanne Graham",
        "username": "Bret",
        "email": "Sincere@april.biz",
        "address": {
            "street": "Kulas Light",
            "suite": "Apt. 556",
            "city": "Gwenborough",
            "zipcode": "92998-3874",
            "geo": {
                "lat": "-37.3159",
                "lng": "81.1496"
            }
        },
        "phone": "1-770-736-8031 x56442",
        "website": "hildegard.org",
        "company": {
            "name": "Romaguera-Crona",
            "catchPhrase": "Multi-layered client-server neural-net",
            "bs": "harness real-time e-markets"
        }
    }
}
path = (Path(__file__) / "..").resolve()

# Data structure: function | (arguments, kwargs) | expected (args, kwargs)
test_convert_args_data = [
    pytest.param(
        convert_args(echo, int, float),
        (("10", "15.5"), {}),
        ((10, 15.5), {}),
        id="base"
    ),
    pytest.param(
        convert_args(echo, int, float),
        (("10", "15.5", "abc", "qwerty"), {}),
        ((10, 15.5, "abc", "qwerty"), {}),
        id="extra_args"
    ),
    pytest.param(
        convert_args(echo, int, float, x=int, y=float),
        (("10", "15.5", "abc", "qwerty"), {'x': '0', 'y': '1.5', 'z': '0'}),
        ((10, 15.5, "abc", "qwerty"), {'x': 0, 'y': 1.5, 'z': '0'}),
        id="kwargs"
    )
]


@pytest.mark.parametrize("fn,value,expected", test_convert_args_data)
def test_convert_args(fn, value, expected):
    args, kwargs = value
    result = fn(*args, **kwargs)
    assert result == expected, f"(in value: {value})"


test_path_split_data = [
    pytest.param("./test/file.txt", ['.', 'test', 'file.txt'], id='base'),
    pytest.param("/tmp/test/file.txt", ['/', 'tmp', 'test', 'file.txt'], id='unix path'),
    pytest.param("test/*.txt", ['test', '*.txt'], id='pattern'),
    pytest.param("test/**/*.txt", ['test', '**', '*.txt'], id='recursive pattern'),
]

if sys.platform.startswith("win"):
    test_path_split_data.append(pytest.param("X:/test/file.txt", ['X:/', 'test', 'file.txt'], id='windows path'))
else:
    test_path_split_data.append(pytest.param("X:/test/file.txt", ['X:', 'test', 'file.txt'], id='windows path'))


@pytest.mark.parametrize("value,expected", test_path_split_data)
def test_path_split(value, expected):
    result = path_split(value)
    assert result == expected, f"(in value: {value})"


test_dict_lookup_data = [
    pytest.param("user.name", test_dict["user"]["name"], id="basic"),
    pytest.param("-", test_dict, id="no lookup"),
    pytest.param("user", test_dict["user"], id="single"),
]


@pytest.mark.parametrize("value,expected", test_dict_lookup_data)
def test_dict_lookup(value, expected):
    result = dict_lookup(test_dict, value)
    assert result == expected, f"(in value: {value})"


test_iter_json_file_data = [
    pytest.param(
        (
            json.load((path / "data" / "users.json").open()),
            "-"
        ),
        lambda data: len(data) == 10
    ),
    pytest.param(
        (
            json.load((path / "data" / "photos.json").open()),
            "items"
        ),
        lambda data: len(data) == 5000
    ),
]


@pytest.mark.parametrize("value,expected", test_iter_json_file_data)
def test_iter_json_file(value, expected):
    result = list(iter_json_file(*value))
    assert expected(result) is True, f"(in value: {value})"


abs_path = path.absolute()
abs_path_str = str(abs_path).replace("\\", "/")
# noinspection PyRedeclaration
test_process_path_data = [
    pytest.param(
        abs_path / "data" / "dummy_files" / "test1.txt",
        {abs_path_str + "/data/dummy_files/test1.txt"},
        id="single_file"
    ),
    pytest.param(
        abs_path / "data" / "dummy_files" / "*.txt",
        {abs_path_str + f"/data/dummy_files/test{i}.txt" for i in range(1, 4)},
        id="pattern"
    ),
    pytest.param(
        abs_path / "data" / "dummy_files" / "*.csv",
        {abs_path_str + f"/data/dummy_files/test{i}.csv" for i in range(1, 3)},
        id="pattern_2"
    ),
    pytest.param(
        abs_path / "data" / "*" / "*",
        {abs_path_str + f"/data/dummy_files/test{i}.txt" for i in range(1, 4)} |
        {abs_path_str + f"/data/dummy_files/test{i}.csv" for i in range(1, 3)},
        id="pattern_3"
    ),
    pytest.param(
        Path("*.py"),
        {p.replace("\\", "/")
         for p in map(str, Path('.').iterdir())
         if p.endswith(".py")},
        id="pattern_4"
    )
]


@pytest.mark.parametrize("value,expected", test_process_path_data)
def test_process_path(value, expected):
    result = set(str(p).replace("\\", "/") for p in _process_path(value))
    assert result == expected, f"(in value: {value})"



from enum import Enum
from typing import Dict, List, Tuple

from ..dynamic_typing import ModelMeta

Index = str
ModelsStructureType = Tuple[List[dict], Dict[ModelMeta, ModelMeta]]

INDENT = " " * 4
OBJECTS_DELIMITER = "\n" * 3  # 2 blank lines


class ClassType(Enum):
    Dataclass = "dataclass"
    Attrs = "attrs"
    Pydantic = "pydantic"



from functools import partial
from itertools import chain
from typing import AbstractSet, Dict, Iterable, List, Optional, Tuple, Type, Union

from typing_extensions import Literal

from .base import BaseType, ImportPathList, MetaData, get_hash_string
from .typing import metadata_to_typing


class SingleType(BaseType):
    _typing_cls = None
    __slots__ = ["_type", "_hash"]

    def __init__(self, t: MetaData):
        self._type = t
        self._hash = None

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, t: MetaData):
        self._type = t
        self._hash = None

    def __str__(self):
        return f"{type(self).__name__}[{self.type}]"

    def __repr__(self):
        return f"<{type(self).__name__} [{self.type}]>"

    def __iter__(self) -> Iterable['MetaData']:
        yield self.type

    def __eq__(self, other):
        return type(other) is type(self) and self.type == other.type

    def replace(self, t: 'MetaData', **kwargs) -> 'SingleType':
        self.type = t
        return self

    def to_typing_code(self, types_style: Dict[Union['BaseType', Type['BaseType']], dict]) \
            -> Tuple[ImportPathList, str]:
        imports, nested = metadata_to_typing(self.type, types_style=types_style)
        return (
            [*imports, (self._typing_cls.__module__, self._typing_cls._name)],
            f"{self._typing_cls._name}[{nested}]"
        )

    def _to_hash_string(self) -> str:
        return f"{type(self).__name__}/{get_hash_string(self.type)}"


class ComplexType(BaseType):
    _typing_cls = None
    __slots__ = ["_types", "_sorted", "_hash"]

    def __init__(self, *types: MetaData):
        self._types = list(types)
        self._sorted = None
        self._hash = None

    @property
    def types(self):
        return self._types

    @types.setter
    def types(self, value):
        self._types = value
        self._sorted = None
        self._hash = None

    @property
    def sorted(self):
        """
        Getter of cached sorted types list
        """
        sorted_types = getattr(self, '_sorted', None)
        if sorted_types is None:
            sorted_types = sorted(self.types, key=self._sort_key)
            self._sorted = sorted_types
        return sorted_types

    @staticmethod
    def _sort_key(item):
        if hasattr(item, 'keys'):
            return str(sorted(item.keys()))
        else:
            return str(item)

    def __str__(self):
        items = ', '.join(map(str, self.types))
        return f"{type(self).__name__}[{items}]"

    def __repr__(self):
        items = ', '.join(map(str, self.types))
        return f"<{type(self).__name__} [{items}]>"

    def __iter__(self) -> Iterable['MetaData']:
        yield from self.types

    def __eq__(self, other):
        return type(other) is type(self) and self.sorted == other.sorted

    def __len__(self):
        return len(self.types)

    def replace(self, t: Union['MetaData', List['MetaData']], index=None, **kwargs) -> 'ComplexType':
        if index is None and isinstance(t, list):
            self.types = t
        elif index is not None and not isinstance(t, list):
            types = self.types
            types[index] = t
            # Using property setter here
            self.types = types
        else:
            raise ValueError(f"Unsupported arguments: t={t} index={index} kwargs={kwargs}")
        return self

    def to_typing_code(self, types_style: Dict[Union['BaseType', Type['BaseType']], dict]) \
            -> Tuple[ImportPathList, str]:
        imports, nested = zip(*map(partial(metadata_to_typing, types_style=types_style), self))
        nested = ", ".join(nested)
        return (
            [*chain.from_iterable(imports), (self._typing_cls.__module__, self._typing_cls._name)],
            f"{self._typing_cls._name}[{nested}]"
        )

    def _to_hash_string(self) -> str:
        return type(self).__name__ + "/" + ",".join(map(get_hash_string, self.types))


class DOptional(SingleType):
    """
    Field of this type may not be presented in JSON object
    """
    _typing_cls = Optional


class DUnion(ComplexType):
    """
    Same as typing.Union. Nested types are unique.
    """
    _typing_cls = Union

    def __init__(self, *types: Union[type, BaseType, dict]):
        hashes = set()
        unique_types = []
        use_literals = True
        str_literals = set()

        # Ensure that types in union are unique

        def handle_type(t, use_literals):
            if t is str:
                use_literals = False

            if isinstance(t, StringLiteral):
                if not use_literals:
                    return

                if t.overflowed:
                    use_literals = False
                else:
                    str_literals.update(t.literals)

            else:
                h = get_hash_string(t)
                if h not in hashes:
                    unique_types.append(t)
                    hashes.add(h)
            return use_literals

        for t in types:
            if isinstance(t, DUnion):
                # Merging nested DUnions
                for t2 in list(t._extract_nested_types()):
                    use_literals = handle_type(t2, use_literals) and use_literals
            else:
                use_literals = handle_type(t, use_literals) and use_literals

        if str_literals and use_literals:
            literal = StringLiteral(str_literals)
            if literal.overflowed:
                use_literals = False
            else:
                unique_types.append(literal)

        if not use_literals:
            handle_type(str, use_literals=False)
        super().__init__(*unique_types)

    def _extract_nested_types(self):
        """
        Same as ComplexType.__iter__ but "flatten" nested DUnions
        """
        for t in self.types:
            if isinstance(t, DUnion):
                yield from t._extract_nested_types()
            else:
                yield t


class DTuple(ComplexType):
    _typing_cls = Tuple


class DList(SingleType):
    _typing_cls = List


class DDict(SingleType):
    _typing_cls = Dict

    # Dict is single type because keys of JSON dict are always strings.
    def to_typing_code(self, types_style: Dict[Union['BaseType', Type['BaseType']], dict]) \
            -> Tuple[ImportPathList, str]:
        imports, nested = metadata_to_typing(self.type, types_style=types_style)
        return (
            [*imports, ('typing', 'Dict')],
            f"Dict[str, {nested}]"
        )


class StringLiteral(BaseType):
    class TypeStyle:
        use_literals = 'use_literals'
        max_literals = 'max_literals'

    MAX_LITERALS = 15  # Hard limit for performance optimization
    MAX_STRING_LENGTH = 20
    __slots__ = ["_literals", "_hash", "_overflow"]

    def __init__(self, literals: AbstractSet[str]):
        self._overflow = (
                len(literals) > self.MAX_LITERALS
                or
                any(map(
                    lambda s: len(s) >= self.MAX_STRING_LENGTH,
                    literals
                ))
        )
        self._literals = frozenset() if self._overflow else literals

    def __iter__(self) -> Iterable['MetaData']:
        return iter(())

    def __str__(self):
        return f"{type(self).__name__}[{self._repr_literals()}]"

    def __repr__(self):
        return f"<{type(self).__name__} [{self._repr_literals()}]>"

    def __eq__(self, other):
        return type(other) is type(self) and self._literals == other._literals

    def replace(self, t: 'MetaData', **kwargs) -> 'StringLiteral':
        return self

    def to_typing_code(self, types_style: Dict[Union['BaseType', Type['BaseType']], dict]) \
            -> Tuple[ImportPathList, str]:
        options = self.get_options_for_type(self, types_style)
        if options.get(self.TypeStyle.use_literals):
            limit = options.get(self.TypeStyle.max_literals)
            if limit is None or len(self.literals) < limit:
                parts = ', '.join(
                    '"{}"'.format(s.replace('\\', '\\\\').replace('"', '\\"'))
                    for s in sorted(self.literals)
                )
                return [(Literal.__module__, 'Literal')], f"Literal[{parts}]"

        return [], 'str'

    def _to_hash_string(self) -> str:
        return f"{type(self).__name__}/{self._repr_literals()}"

    @property
    def literals(self):
        return self._literals

    @property
    def overflowed(self):
        return self._overflow

    def _repr_literals(self):
        if self._overflow:
            return '...'
        return ','.join(self._literals)



import operator
from datetime import date, datetime, time
from inspect import isclass
from typing import Dict, Set, Tuple, Type, Union

from .base import BaseType, ImportPathList, MetaData
from .string_serializable import StringSerializable


def metadata_to_typing(t: MetaData, types_style: Dict[Union[BaseType, Type[BaseType]], dict] = None) \
        -> Tuple[ImportPathList, str]:
    """
    Shortcut function to call ``to_typing_code`` method of BaseType instances or return name of type otherwise
    :param t:
    :return:
    """
    types_style = types_style or {}
    if isclass(t):
        if issubclass(t, StringSerializable):
            return t.to_typing_code(types_style)
        else:
            imports = []
            if issubclass(t, (date, datetime, time)):
                imports.append((t.__module__, [t.__name__]))
            return (imports, t.__name__)
    elif isinstance(t, dict):
        raise ValueError("Can not convert dict instance to typing code. It should be wrapped into ModelMeta instance")
    else:
        return t.to_typing_code(types_style)


def compile_imports(imports: ImportPathList) -> str:
    """
    Merge list of imports path and convert them into list code (string)
    """
    class_imports_map: Dict[str, Set[str]] = {}
    package_imports_set: Set[str] = set()
    for module, classes in filter(None, imports):
        if classes is None:
            package_imports_set.add(module)
        else:
            classes_set = class_imports_map.get(module, set())
            if isinstance(classes, str):
                classes_set.add(classes)
            else:
                classes_set.update(classes)
            class_imports_map[module] = classes_set

    # Sort imports by package name and sort class names of each import
    class_imports_map = dict(sorted(
        ((module, sorted(classes)) for module, classes in class_imports_map.items()),
        key=operator.itemgetter(0)
    ))

    class_imports = "\n".join(
        f"from {module} import {', '.join(classes)}"
        for module, classes in class_imports_map.items()
    )
    package_imports = "\n".join(
        f"import {module}"
        for module in sorted(package_imports_set)
    )
    return "\n".join(filter(None, (package_imports, class_imports)))
